!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: btreebasicinter.gs,v 1.7 2008-01-09 19:02:29 stever Exp $
!
! Superclass Hierarchy:
!   BtreeBasicInteriorNode, BtreeInteriorNode, BtreeNode, Array,
!   SequenceableCollection, Collection, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods BtreeBasicInteriorNode
removeallclassmethods BtreeBasicInteriorNode

protectmethods

category: 'For Documentation Installation only'
classmethod: BtreeBasicInteriorNode
installDocumentation

< unprotected >

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
'A BtreeBasicInteriorNode represents an interior node of a B-tree for which a
 binary search can be done for a key.' .
doc documentClassWith: txt.

self description: doc.
%

category: 'Searching'
method: BtreeBasicInteriorNode
_binarySearchCoveringKey: aKey totalOrder: aBoolean 

"Returns the index for the first entry in which aKey is found utilizing a
 binary search.  This is the first entry whose key >= aKey."

<protected primitive: 206>
self _primitiveFailed: #_binarySearchCoveringKey:totalOrder: .
self _uncontinuableError
%

category: 'Searching'
method: BtreeBasicInteriorNode
_binarySearchCoveringKey: aKey value: aValue 

"Returns the index for the entry in which aKey/aValue would be
 inserted.  This is the first entry which has its key > aKey.  If
 multiple entries have the same key, then aValue is used (since
 entries with duplicate keys are inserted by the value's OOP ordering).
 If the receiver is empty, returns 1.  Uses a binary search."

<protected primitive: 85>
self _primitiveFailed: #_binarySearchCoveringKey:value: .
self _uncontinuableError
%

category: 'Searching'
method: BtreeBasicInteriorNode
_binarySearchForKey: aKey value: aValue 

"Returns the 'value' index whose key equality matches aKey.
 If no match is found, returns zero.  Uses a binary search."

<protected primitive: 113>
self _primitiveFailed: #_binarySearchForKey:value: .
self _uncontinuableError
%

category: 'Constants'
method: BtreeBasicInteriorNode
_encryptionSize

"Three slots are used for encryption."
<protected>

^ 3
%

category: 'Constants'
method: BtreeBasicInteriorNode
entrySize

"Returns the size of an entry (with 3 slots for encryption)."
<protected>

^ 5
%

category: 'Constants'
classmethod: BtreeBasicInteriorNode
entrySize

"Returns the size of an entry (with 3 slots for encryption)."
<protected>

^ 5
%

category: 'Constants'
classmethod: BtreeBasicInteriorNode
maxNumberOfElements

"Returns the number of entries that are allowed in a leaf node."
<protected>

^ 400
%

category: 'Constants'
method: BtreeBasicInteriorNode
mergeThreshold

"Returns the threshold such that if an entry has less than this number, it must
 have elements added from its sibling or be merged with its sibling.  This
 amount is usually the same as the result of the message
 'self maxNumberOfElements quo: 2'."
<protected>

^ 200
%

category: 'Constants'
method: BtreeBasicInteriorNode
parentNodeClass

"Returns the class of node to be created as the parent when a split occurs."
<protected>

^ BtreeBasicInteriorNode
%

category: 'Testing'
method: BtreeBasicInteriorNode
_canCompare: aKey withClass: aClass

"Returns true if the receiver can make comparisons with the given key, and
 false otherwise."
<protected>

^ self _basicCanCompare: aKey withClass: aClass
%

category: 'Constants'
method: BtreeBasicInteriorNode
_emptyEntryArray

"Returns a five-element Array of nils."
<protected>

^ #( nil nil nil nil nil)
%

category: 'Accessing'
method: BtreeBasicInteriorNode
_lastIndex

"Returns the index of the last used slot in the node."
<protected>

^ numElements * 5
%

category: 'Accessing'
method: BtreeBasicInteriorNode
_lastKeyIndex

"Returns the index of the key of the last entry, or zero if the node is empty."
<protected>

^ numElements == 0
    ifTrue: [ 0 ]
    ifFalse: [ numElements * 5 - 3 ]
%

category: 'Comparison Operators'
method: BtreeBasicInteriorNode
_compareKey: aKey equalToEntryAt: index 

"Performs an equality comparison between aKey and the entry whose key is at the
 given index.  Use encryption to make the comparison faster."

<protected primitive: 40>
self _primitiveFailed: #_compareKey:equalToEntryAt: .
self _uncontinuableError
%

category: 'Testing'
method: BtreeBasicInteriorNode
_isSymbolAt: offset

"Returns true if the key at the given offset is a Symbol; returns false
 otherwise.  This method uses the encryption to avoid fetching the object."

<protected primitive: 58>
self _primitiveFailed: #_isSymbolAt: .
self _uncontinuableError
%

unprotectmethods

