!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: clusterb.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   ClusterBucket, Object.
!
!=========================================================================

removeallmethods ClusterBucket
removeallclassmethods ClusterBucket

category: 'For Documentation Installation only'
classmethod: ClusterBucket
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A ClusterBucket describes clustering behavior, such as the identity of the
 extent to which clustered objects are to be written.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A SmallInteger.'.
doc documentInstVar: #_clusterId with: txt.

txt := (GsDocText new) details:
'A SmallInteger.'.
doc documentInstVar: #extentId with: txt.

txt := (GsDocText new) details:
'A Boolean.'.
doc documentInstVar: #keepClusteredOnModify with: txt.

txt := (GsDocText new) details:
'Unconstrained, typically a String.'.
doc documentInstVar: #description with: txt.

self description: doc.
%

category: 'Accessing'
classmethod: ClusterBucket
allInstances

"Returns the collection of all instances of the receiver."

^ AllClusterBuckets
%

category: 'Accessing'
classmethod: ClusterBucket
bucketWithId: aPositiveSmallInt

"Returns the instance with the specified ID if one exists.  Generates an error
 if anInt is less than 1 or outside of the range of existing cluster buckets."

^ AllClusterBuckets at: aPositiveSmallInt
%

category: 'Instance Creation'
classmethod: ClusterBucket
new

"Creates an instance of the receiver and adds the new instance to
 AllClusterBuckets.  Requires write authorization to the DataCurator segment."

| newId result |
(AllClusterBuckets size >= 67108863) ifTrue: [
  "constant 67108863 related to OBJ_CLUSTER_ID_MASK in virtual machine"
  self _error: #rtErrMaxClusterId args: 67108863.
  ^ AllClusterBuckets at: 1 "returns default bucket"
  ] .
newId := AllClusterBuckets size + 1 .
result := super new .
result _clusterId: newId ;
       extentId: nil "attempt to keep objects in their current extent" ;
       keepClusteredOnModify: false ;
  "create an empty string for the description, so description can be
   modified later without putting the ClusterBucket in a write set."
       description: String new .

AllClusterBuckets addLast: result .
^ result
%

category: 'Instance Creation'
classmethod: ClusterBucket
newForExtent: extentId

"Creates an instance of the receiver for clustering objects in the extent
 extentId.  The extentId argument is a positive SmallInteger in the range of
 1 to (SystemRepository numberOfExtents).  Requires write authorization to the
 DataCurator segment."

^ self new extentId: extentId
%

category: 'Accessing'
method: ClusterBucket
clusterId

"Returns the value of the private instance variable.  This instance variable
 should only be assigned by ClusterBucket | new."

^ _clusterId
%

category: 'Accessing'
method: ClusterBucket
extentId

"Returns the value of the extentId instance variable."

^ extentId
%

category: 'Accessing'
method: ClusterBucket
keepClusteredOnModify

"This feature is not implemented in this release."

^ keepClusteredOnModify
%

category: 'Accessing'
method: ClusterBucket
description

"Returns the value of the description instance variable."

^ description
%

category: 'Updating'
method: ClusterBucket
description: anObject

"Assigns anObject (typically some kind of String object) as the description of
 the receiver."

description := anObject
%

category: 'Updating'
method: ClusterBucket
_clearClusterCache

"Private.  Any method that modifies an instance variable of a
 ClusterBucket that affects clustering behavior must send this method.  The
 instance variable description is not included in such instance variables."

<primitive: 235>
self _primitiveFailed: #_clearClusterCache .
self _uncontinuableError
%

category: 'Updating'
method: ClusterBucket
_clusterId: aSmallInteger

"Private.  For use only during instance creation."

_clusterId := aSmallInteger .
self _clearClusterCache .
%

category: 'Updating'
method: ClusterBucket
extentId: anExtentId

"An argument of nil specifies don't care behavior.  The bucket uses the next
 available disk page from any extent based on the allocation mode defined in
 the Stone's configuration file.  Positive arguments are an offset into the
 result of Repository | fileNames, thus specifying an extent.  Reference to an
 extent which does not exist will generate an error at the time of executing
 this method.

 At the time of object modification, a non-nil extentId specifies which extent
 to put the object in.  If the extent no longer exists, or is full, 
 at the time of clustering or object modification, don't care behavior occurs."

SystemRepository validateExtentId: anExtentId .
extentId := anExtentId .
self _clearClusterCache .
%

category: 'Updating'
method: ClusterBucket
keepClusteredOnModify: aBoolean

"This feature is deferred until a future release.  Argument value of true
 is not supported in this release."

keepClusteredOnModify := false .
self _clearClusterCache .
%

category: 'Accessing the Class Format'
classmethod: ClusterBucket
firstPublicInstVar

"Returns the index of the first publicly available instance variable storage
 location, whether or not a public instance variable has actually been defined."

^ 2
%

