!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: compile.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   CompiledMethod, Object.
!
!=========================================================================

removeallmethods CompiledMethod
removeallclassmethods CompiledMethod

category: 'For Documentation Installation only'
classmethod: CompiledMethod
installDocumentation

| doc txt |

doc := GsClassDocumentation _newForObsoleteGsClass: self asOfGsVersion: '5.0'.

txt := (GsDocText new) details:
'CompiledMethod has been replaced by GsMethod.' .
doc documentClassWith: txt.

self description: doc.
%

category: 'Debugging Support'
classmethod: CompiledMethod
_sourceWithErrors: compilerError fromString: aString

"This method returns an instance of aString's class containing the text in a
 string with compiler errors marked, plus the error text for each error."

| lineFeed result aStringSize offsets errNumbers thisErr pos
  markerArray errDict |

"initialize"
lineFeed := Character lf.
"only ask for source size once for efficiency for Japanese"
aStringSize:= aString size.
offsets := Array new: compilerError size.
errNumbers := Array new: compilerError size.

"get an Array of source offsets with errors, and an Array of error numbers"
1 to: compilerError size do: [:i |
   thisErr := compilerError at: i.
   offsets at: i put: (thisErr at: 2 "source offset").
   errNumbers at: i put: (thisErr at: 1 "error number")
].

"build an Array parallel to the source that contains nil if no error at
 that source position, and an index into offsets if there is an error at
 that source position"
markerArray:= self _buildMarkersFrom: offsets ofSize: aStringSize.

result:= self _buildMarkedSourceFrom: aString
                          sourceSize: aStringSize
                             markers: markerArray.

"add error strings"
errDict := GemStoneError at: (System myUserProfile nativeLanguage).
1 to: errNumbers size do: [:i |
  result add: lineFeed.
  result addAll: i asString.
  result addAll: ': ['.
  pos := errNumbers at: i.
  result addAll: pos asString.
  result addAll: '] '.
  pos > errDict size
    ifTrue: [ result addAll: ' (illegal)']
    ifFalse: [ result addAll: (errDict at: pos) asString]
  ].
result add: lineFeed.

^result
%

category: 'Debugging Support'
classmethod: CompiledMethod
_buildMarkedSourceFrom: sourceStr sourceSize: aSize markers: markerArray

"Given a source string, its size (passed in for efficiency), and a marker
 Array, returns an instance of sourceStr's class containing the marked source."

| lineFeed tab space
  errorMarked     "an Array of errors marked on the current line"
  markerLineIndex "index into the current marker line"
  result          "the result of this method"
  markerLine      "the current marker line"
  aChar           "one Character of the source"
  displayWidth    "the number of positions it takes to display aChar"
|

"initialize"
lineFeed := Character lf.
tab:= Character tab.
space:= Character space.

errorMarked:= Array new.
markerLineIndex:= 1.
result:= sourceStr class new .
result addAll: '   '.
markerLine:= String new markerLine add: $ ; yourself .
1 to: aSize do: [:i |
   aChar:= sourceStr at: i.  "fetch a char"
   displayWidth:= aChar displayWidth.

   "Add the char to the result"
   (aChar = tab)
   ifTrue: [
      displayWidth:= 8 - ((i - 1) \\ 8).
      displayWidth timesRepeat: [result add: space].
   ]
   ifFalse: [
      result add: aChar.
      ((i == aSize) _and: [aChar ~~ lineFeed])
         ifTrue: [result add: lineFeed].
   ].

   ((markerArray at: i) == nil) "no error at this position"
   ifTrue: [
      displayWidth timesRepeat: [markerLine add: space].
   ]
   ifFalse: [ "found an error at this position"
      errorMarked add: #[markerLineIndex + 1, markerArray at: i].
      markerLine add: $^.
      displayWidth - 1 timesRepeat: [markerLine add: space].
   ].
   markerLineIndex:= markerLineIndex + displayWidth.

   ((aChar = lineFeed) _or: [i == aSize])
   ifTrue: [ "we are at end of line"

      "add error identifiers to marker line "
      (errorMarked size ~~ 0) ifTrue: [
         self _addErrorIds: #[errorMarked, markerLine, markerLineIndex].
         result add: $  ; add: $* .
         result add: markerLine.
         ] .

      (i == aSize) ifFalse: [result addAll: '   ' ].
      markerLine size: 1.
      markerLineIndex:= 1.
      errorMarked size: 0.
   ]
].
^result
%


category: 'Debugging Support'
classmethod: CompiledMethod
_buildMarkersFrom: anArray ofSize: aSize

"Given an Array of offsets, build an Array of size aSize containing
 the index into anArray at the position corresponding to anArray's element.
 The remainder of the Array contains nil."

| markerArray offset |
markerArray:= Array new: aSize.
1 to: anArray size do: [:i |
   offset:= ((anArray at: i) max: 1) min: aSize.  "limit within range"
   ((markerArray at: offset) == nil)
   ifTrue: [ markerArray at: offset put: i]
].
^markerArray
%

category: 'Debugging Support'
classmethod: CompiledMethod
_addErrorIds: anArray

""

| errorMarked markerLine markerLineSize space addToEnd errorPosition
  aStr neededSize subStr |

errorMarked:= anArray at: 1.
markerLine:= anArray at: 2.
markerLineSize:= anArray at: 3.
space:= Character space.

"have the source marked at each error with ^; now add error identifier"
addToEnd:= false.
1 to: (errorMarked size) do: [:i |
   errorPosition:= (errorMarked at: i) at: 1.
   aStr:= ((errorMarked at: i) at: 2) asString.
   neededSize:= errorPosition + aStr size.
   (markerLineSize < neededSize)
   ifTrue: [
       markerLine size: neededSize.
       markerLineSize + 1 to: neededSize do: [:i |
          markerLine at: i put: space].
       markerLineSize:= neededSize.
   ].

   (addToEnd)
   ifFalse: [
      subStr:= markerLine copyFrom: errorPosition + 1
                                to: (errorPosition + aStr size).
      subStr do: [:each | (each = $ ) ifFalse: [addToEnd:= true]].
   ].
   (addToEnd)
   ifTrue: [ markerLine add: aStr.
             (i == errorMarked size)
             ifFalse: [ markerLine add: ',']]
   ifFalse: [ aStr copyFrom: 1 to: (aStr size) into: markerLine
                   startingAt: errorPosition + 1.]
].
(68 - markerLine size) timesRepeat:[ markerLine add: $ ].
(75 - markerLine size) timesRepeat:[ markerLine add: $* ] .
markerLine add: Character lf.
^ true
%

category: 'Instance Creation'
classmethod: CompiledMethod
new: anInteger

"Disallowed.  You cannot create new instances of CompiledMethod."

self shouldNotImplement: #new:
%

category: 'Instance Creation'
classmethod: CompiledMethod
new

"Disallowed.  You cannot create new instances of CompiledMethod."

self shouldNotImplement: #new
%

category: 'Debugging Support'
method: CompiledMethod
_buildMarkerArray: allSteps ofSize: aSize

"This method builds a marker Array for the receiver's source code string.
 allSteps is true if all steps are to be marked.  It is false if only steps
 that correspond to set breakpoints need be marked.  The Array is the same
 size as the source string and contains step numbers at offsets corresponding
 to the source string."

| markerArray allBreaks breakArray |

markerArray := Array new: aSize.
allSteps ifTrue:
  [^ CompiledMethod _buildMarkersFrom: sourceOffsets ofSize: aSize].

breakArray := System _breakpointReport at: 3.
breakArray == nil ifTrue:
  [ ^ markerArray ].

allBreaks := IdentitySet new.
breakArray do:
  [ :each |
    (each ~~ nil _and: [ (each at: 2) == self ])
    ifTrue:
      [ allBreaks add: (each at: 3) ].
    ].
allBreaks size == 0 ifTrue:
    [ ^ markerArray ].

1 to: sourceOffsets size do:
  [ :i |
    (allSteps _or: [ i in: allBreaks ])
    ifTrue:
      [ markerArray at: (sourceOffsets at: i) put: i.  ].
  ].

^ markerArray
%

category: 'Debugging Support'
method: CompiledMethod
_sourceWithSteps: allSteps

"This method returns the source string with intermixed control information
 indicating where step points are.  allSteps indicates whether all steps should
 be indicated or only those steps where a breakpoint currently exists."

| markerArray "A parallel Array to the source string.  It is filled with nils
               except for locations pointed to by the sourceOffsets Array."
  aSize "the size of the source"
|

aSize:= sourceString size.
  "only ask for the source size once for efficiency for Japanese"
markerArray := self _buildMarkerArray: allSteps ofSize: aSize.
^ CompiledMethod _buildMarkedSourceFrom: sourceString
                             sourceSize: aSize
                                markers: markerArray
%

category: 'Debugging Support'
method: CompiledMethod
_sourceOffsets

"Returns an InvariantArray (that holds SmallIntegers) which is a list
 of offsets into sourceString, corresponding in order to the step points."

^ sourceOffsets
%

category: 'Debugging Support'
method: CompiledMethod
_breakPointKind: anIp

"This method infers the kind of action associated with a given bytecode."

| theByteCode |

theByteCode := (byteCodes at: anIp + 1) asciiValue.
(theByteCode >= 12 _and: [theByteCode <= 25]) ifTrue: [^ #store].
(((theByteCode >= 26 _and: [theByteCode <= 31]) _or:
  [theByteCode == 47]) _or: [theByteCode == 48]) ifTrue: [^ #send].
(theByteCode >= 42 _and: [theByteCode <= 44]) ifTrue: [^ #return].
^ #unknown
%

category: 'Debugging Support'
method: CompiledMethod
_stepPointForIp: anIp

"Returns the offset in the ipSteps Array of the receiver which represents
 the step point at or just after the given instruction pointer offset"

"This method is useful when doing low-level debugging of the virtual machine."

1 to: ipSteps size do:[:j |
  (ipSteps at:j) >= anIp ifTrue:[ ^ j ]
  ].
^ ipSteps size
%

category: 'Debugging Support'
method: CompiledMethod
_sendCount: anIp

"This method returns the number of arguments (excluding receiver)
 associated with the message send at this offset."

| theByteCode specialCode|

theByteCode := (byteCodes at: anIp + 1) asciiValue.
(theByteCode = 26) ifTrue: [^ 0].  " BC_SEND_NO_ARGS "
(theByteCode = 27) ifTrue: [^ 1].  "BC_SEND_ONE_ARG"
(theByteCode = 28) ifTrue: [^ 2].  "BC_SEND_TWO_ARGS"
(theByteCode == 29 "BC_SEND_WITH_ARGS"
 _or: [theByteCode == 30] "BC_SEND_SUPER_WITH_ARGS" )
   ifTrue: [^ (byteCodes at: anIp + 2) asciiValue ].

theByteCode == 31 "BC_SEND_SPECIAL"
ifTrue: [  "Special send, lots of work"
  specialCode := (byteCodes at: anIp + 2) asciiValue.
  specialCode <= 13 ifTrue: [^ 1].  "binary selector"
  specialCode == 14 ifTrue: [^ 2].  "at:put:"
  specialCode == 15 ifTrue: [^ 1].  "at:"
  specialCode == 16 ifTrue: [^ 0].  "class"
  specialCode == 17 ifTrue: [^ 1].  "isKindOf:"
  specialCode == 18 ifTrue: [^ 0].  "size"
  ].
^ 0 "who knows?"
%

category: 'Accessing'
method: CompiledMethod
_blockPool

"Returns the value of the instance variable blockPool."

^ blockPool
%

category: 'Accessing'
method: CompiledMethod
_requiresVC

"Returns the value of the instance variable requiresVC."

^ requiresVC
%

category: 'Accessing'
method: CompiledMethod
_argsAndTemps

"Returns the value of the instance variable argsAndTemps."

^ argsAndTemps
%

category: 'Accessing'
method: CompiledMethod
_byteCodes

"Returns the value of the instance variable byteCodes."

^ byteCodes
%

category: 'Accessing'
method: CompiledMethod
_stackSize

"Returns the value of the instance variable stackSize."

^ stackSize
%

category: 'Accessing'
method: CompiledMethod
at: anIndex put: anObject

"Disallowed.  You may not change the value of a CompiledMethod."

self shouldNotImplement: #at:put:
%

category: 'Accessing'
method: CompiledMethod
instVarAt: anIndex put: aValue

"Disallowed.  You may not change the value of a CompiledMethod."

self shouldNotImplement: #instVarAt:put:
%

category: 'Accessing'
method: CompiledMethod
size: anInteger

"Disallowed.  You may not change the size of a CompiledMethod."

self shouldNotImplement: #size:
%

category: 'Accessing'
method: CompiledMethod
_inClass

"Returns the value of the instance variable inClass."

^ inClass
%

category: 'Accessing'
method: CompiledMethod
_ipSteps

""

^ ipSteps
%

category: 'Accessing'
method: CompiledMethod
_index

"Returns the value of the instance variable index."

^ index
%

category: 'Accessing'
method: CompiledMethod
_methodType

"Returns the value of the instance variable methodType."

^ methodType
%

category: 'Accessing'
method: CompiledMethod
_nArgs

"Returns the value of the instance variable nargs."

^ nArgs
%

category: 'Accessing'
method: CompiledMethod
_selector

"Returns the value of the instance variable named selector."

^ selector
%

category: 'Accessing'
method: CompiledMethod
_selectorPool

"Returns the value of the instance variable selectorPool."

^ selectorPool
%

category: 'Accessing'
method: CompiledMethod
_sourceString

"Returns the value of the instance variable sourceString."

| result |
result := String new.
result add: ' '.
sourceString copyFrom: 1 to: sourceString size
             into: result startingAt: 2.
^ result
%

category: 'Accessing'
method: CompiledMethod
_stackBase

"Returns the value of the instance variable stackBase."

^ stackBase
%

category: 'Clustering'
method: CompiledMethod
clusterDepthFirst

"This method clusters the receiver, its bytecodes, its selector pool, and its
 selector in depth-first order.  Returns true if the receiver has already been
 clustered during the current transaction; returns false otherwise."

| savedBucket |

savedBucket := System clusterBucket.
System clusterBucket: 6.  "kernel miscellany bucket"
argsAndTemps cluster.
(argsAndTemps ~~ nil)
ifTrue:
   [argsAndTemps do: [:each| each cluster]].
blockIps cluster.
sourceOffsets cluster.
ipSteps cluster.
sourceString cluster.
System clusterBucket: savedBucket.

self cluster
ifTrue:
  [ ^ true ]
ifFalse:
  [ byteCodes cluster.
    selectorPool clusterDepthFirst.
    selector cluster.
    1 to: self size do: "get the literals"
      [ :i |
        ((self at: i) isKindOf: Association)
        ifFalse:
          [ (self at: i) clusterDepthFirst ].
      ].
    ^ false
  ]
%

category: 'Copying'
method: CompiledMethod
copy

"Disallowed.  You may not create new instances of CompiledMethod."

self shouldNotImplement: #copy
%

category: 'Private Methods for Class Modification'
method: CompiledMethod
_copyForRecompilation

""

^ super copy
%

category: 'Private Methods for Class Modification'
method: CompiledMethod
_copyToForceRecompilation

"Returns a copy of the receiver, with bytecodes that will force
 recompilation of the receiver prior to subsequent execution."

| newMethod |
newMethod := (CompiledMethod compiledMethodAt: #_methodNeedingRecompilation)
                _copyForRecompilation .
newMethod _sourceString: self _sourceString .
^ newMethod
%

category: 'Error Handling'
method: CompiledMethod
_methodNeedingRecompilation

"A method needs to be recompiled"

"The byteCodes for this method are installed as the byteCodes
 for other methods during schema modification to ensure that
 those methods are recompiled before they are executed."

self _errorUncompiledMethod .
self _uncontinuableError
%

category: 'Private Methods for Class Modification'
method: CompiledMethod
_sourceString: anObj

""

sourceString := anObj
%

category: 'Repository Conversion'
method: CompiledMethod
_recompileWithSymList: aSymbolList category: categorySymbol
  conversionDict: conversionDict 

""

| result newMethod modifyInfo |

result := inClass _primitiveCompileMethod: sourceString
   symbolList: aSymbolList
   category: categorySymbol
   obsoleteClassNames: conversionDict 
   oldLitVars: self _oldLiteralVars  .

result _class == GsMethod ifFalse: [
  newMethod := result at: 1 .
  newMethod _class == GsMethod ifTrue:[
    "compilation succeeded, modify source if needed"
    (result size > 3) ifTrue: [
       (modifyInfo := result at: 4) size > 0 ifTrue:[
          newMethod _modifySourceWith: modifyInfo 
                             oldSource: sourceString 
          ].
       ].
    ].
  ].
^ result.
%

category: 'Repository Conversion'
method: CompiledMethod
_oldLiteralVars

"Returns an Array containing Associations from the literal pool of the
 receiver."

| result |

result := Array new.
1 to: self size do: [:i | | aLit |
  aLit := self at: i .
  (aLit isKindOf: Association) ifTrue:[
    result add: aLit .
    ].
  ].
^ result
%

category: 'Accessing'
method: CompiledMethod
literals

"Returns an OrderedCollection containing the literal pool of the receiver."

| y |

y := OrderedCollection new.
1 to: self size do: [:i |
  y add: (self at: i)
].
^y
%

category: 'Execution'
method: CompiledMethod
_executeInContext: anObject

"The receiver should be a zero argument CompiledMethod obtained by invoking
 String/_compileInContext:symbolList: and the argument should be the argument
 passed to the _compileInContext: keyword for that compilation.

 Returns the result of executing the receiver."

"old CompiledMethod's not executable with new virtual machine"
^ self _primitiveFailed: #_executeInContext:
%

category: 'Stripping Sources'
method: CompiledMethod
_removeAllSourceButFirstComment

"Installs a new source string for the receiver so that only the method
 signature and the first comment (if it exists) are left."

self _unsafeAt: 3 put: self _sourceToFirstComment
%

category: 'Stripping Sources'
method: CompiledMethod
_emptySource

""

^ nil "The string 'source not available...' will be substituted at reload"
%

category: 'Stripping Sources'
method: CompiledMethod
_sourceToFirstComment

"Returns a new source string for the receiver that contains only the method
 signature and the first comment (if it exists).  Does not modify the
 receiver."

| i tmpString sz |
i := 0.
sz := sourceString size.
(selector occurrencesOf: $:) timesRepeat: [
    i := sourceString indexOf: $: startingAt: i + 1.
].
" i is the offset of the last colon in the signature, or zero "
i := i + 1.
" scan past any white space "
[ (sourceString at: i) isSeparator ] whileTrue: [ i := i + 1 ].
" scan past any non-white space "
[ (sourceString at: i) isSeparator not ] whileTrue: [ i := i + 1 ].
" i is now the offset of the first white space past the signature "
" scan past any white space "
[ i <= sz _and: [ (sourceString at: i) isSeparator ] ] whileTrue: [ i := i + 1 ].
" i is now the offset of the initial comment or first line of code "

i > sz
  ifTrue: [ ^ sourceString copy immediateInvariant ].

" if i is the offset of the initial comment, jump to the end of the comment "
(sourceString at: i) == $"
  ifTrue: [ i := sourceString indexOf: $" startingAt: i + 1 ]
  ifFalse: [ i := i - 1 ].

" create the string to replace the original source with "
tmpString := InvariantString new: i .
sourceString copyFrom: 1 to: i into: tmpString startingAt: 1.

tmpString addAll: '

< source code not available >'.

^ tmpString
%

category: 'Decompiling without Sources'
method: CompiledMethod
_bcArrayString

"Returns a string representation of the bytecodes Array."

 | result LF |
result := String new .
LF := Character lf .
result addAll: '#( ' .
1 to: byteCodes size do:[:j|
  j \\ 25 == 0 ifTrue:[ result add: LF ; addAll: '   ' ].
  result addAll:( (byteCodes at:j) asciiValue asString ) .
  result add: $  .  "add a space"
  ].

result addAll:' ) ' .
^ result
%

category: 'Decompiling without Sources'
method: CompiledMethod
_blockPoolArrayString

"Returns an Array describing ExecutableBlocks for a blockPool."

 | result lastSize |
blockPool == nil
  ifTrue:[ result := 'nil' ]
  ifFalse:[
    result := String new .
    result addAll: '#[ ' .
    lastSize := result size .
    1 to: blockPool size do:[:j|
      result addAll: (blockPool at: j) _strippedSource .
      lastSize := result size .
      result addAll: ' , ' .
      ].
    result size: lastSize .
    result addAll:' ] ' .
    ].
^ result

%

category: 'Decompiling without Sources'
method: CompiledMethod
_litArrayString

"Returns a String containing an Array-builder production describing the literal
 pool of the receiver."

| result lastSize |
self size == 0 ifTrue:[ ^ ' #() ' ] .

result := String new .
result addAll: '#[ ' .
lastSize := result size .
1 to: self size do:[:j| | aLit |
  aLit := self at: j .
  (aLit isKindOf: SymbolAssociation)
    ifTrue:[ result addAll: 'System _resolveSym: ' ;
                    addAll: aLit key _asSource
           ]
    ifFalse:[ result addAll: aLit _asSource ] .
  lastSize := result size .
  result addAll: ' ,' ; add: Character lf ; addAll: '   '.
  ].
result size: lastSize .  "remove the last $, "
result addAll:' ] ' .
^ result
%

category: 'Decompiling without Sources'
method: CompiledMethod
_decompileForCategory: aCategory
classRef: classRefExpression
stripWith: sourceStripSelector
classMethod: isMeta

"Decompiles the receiver to produce a topaz run command that will
 regenerate it."


| LF result theClass |

LF := Character lf .

isMeta ifTrue:[ theClass := inClass class ]
      ifFalse:[ theClass := inClass ] .

result := String withAll:'expectvalue %CompiledMethod' .
result add: LF .
result addAll: 'run ' ; add: LF .

result
  addAll: 'CompiledMethod ' ;
  addAll: ' _newInClass: ' ;
      add: '( ' ; addAll: classRefExpression ; addAll: ' )' ; add: LF ;
  addAll: ' classMethod: ' ; addAll: isMeta _asSource ; add: LF ;
  addAll: ' category: ' ; addAll: aCategory _asSource ; add: LF ;
  addAll: ' selector: ' ; addAll: selector _asSource ; add: LF ;

  addAll: ' byteCodes: ' ; addAll: self _bcArrayString ; add: LF ;
  addAll: ' source: ' ;
     addAll: (self perform: sourceStripSelector) _asSource ; add: LF ;

  addAll: ' selectorPool: ' ; addAll: selectorPool _asSource ; add: LF ;
  addAll: ' type: ' ; addAll: methodType asString ;
  addAll: ' index: ' ; addAll: index asString ;
  addAll: ' nargs: ' ; addAll: nArgs asString ;
  addAll: ' sBase: ' ; addAll: stackBase asString ;
  addAll: ' sSize: ' ; addAll: stackSize asString ; add: LF ;
  "ipSteps, sourceOffsets will be regenerated as empty Arrays"
  "argsAndTemps will be regenerated as nil "
  addAll: ' blockIps: ' ; addAll: blockIps _asSource ; add: LF ;
  addAll: ' lang:  ' ; addAll: language asString ;
  addAll: ' reqVC: ' ; addAll: requiresVC asString ; add: LF ;
  addAll: ' blockPool: ' ; addAll: self _blockPoolArrayString ; add: LF ;
  addAll: ' literals: ' ; addAll: self _litArrayString.
                                               "indexed instance variables"

result add: LF .
result add: $% .
result add: LF .
^ result
%

category: 'Reloading Decompiled Methods'
classmethod: CompiledMethod
_newInClass: aClass
classMethod: classMethBool
category: aCategory
selector: aSelector
byteCodes: bcArray
source: partialSource
selectorPool: selectorPoolArray
type: mType
index: idx
nargs: nargs
sBase: sBase
sSize: sSize
blockIps: blockIpsArray
lang: lang
reqVC: reqVC
blockPool: blockPoolArray
literals: litArray

"Private.  Used to regenerate a method from decompiled form."

|cm theClass |
cm := self _new .
classMethBool ifTrue:[ theClass := aClass class ]
             ifFalse:[ theClass := aClass ].

cm _initializeInClass: theClass
   category: aCategory
   selector: aSelector
   byteCodes: bcArray
   source: partialSource
   selectorPool: selectorPoolArray
   type: mType
   index: idx
   nargs: nargs
   sBase: sBase
   sSize: sSize
   blockIps: blockIpsArray
   lang: lang
   reqVC: reqVC
   blockPool: blockPoolArray
   literals: litArray .
^ cm
%

category: 'Reloading Decompiled Methods'
method: CompiledMethod
_initializeInClass: aClass
category: aCategory
selector: aSelector
byteCodes: bcArray
source: partialSource
selectorPool: selectorPoolArray
type: mType
index: idx
nargs: nargs
sBase: sBase
sSize: sSize
blockIps: blockIpsArray
lang: lang
reqVC: reqVC
blockPool: blockPoolArray
literals: litArray

"Private.  Used to regenerate a method from decompiled form."

| byteCodesArray bcSize |

bcSize := bcArray size .
byteCodesArray := InvariantString new: bcSize .
1 to: bcSize do:[:j |
  byteCodesArray at: j put: (Character withValue: (bcArray at: j))
  ].
byteCodes := byteCodesArray .

inClass := aClass .

partialSource == nil
  ifTrue:[  sourceString := '"source not available for this method"' ]
  ifFalse:[ sourceString := partialSource ] .

selectorPool := selectorPoolArray .
methodType := mType .
index := idx .
nArgs := nargs .
stackBase := sBase .
stackSize := sSize .
selector := aSelector .
ipSteps :=  #()  .  "regenerate as empty "
sourceOffsets :=  #()  .  "regenerate as empty"
argsAndTemps := nil .  "regenerate as empty"
blockIps := blockIpsArray .
language := lang .
requiresVC := reqVC .

blockPoolArray == nil
  ifTrue:[ blockPool := blockPoolArray ]
  ifFalse:[
    blockPool := InvariantArray withAll: blockPoolArray .
    blockPool do:[:aBlk | aBlk _method: self ] .  "fix up references to method"
    ].
1 to: litArray size do:[:j |
  self _basicAt: j put: (litArray at: j)
  ].
aClass _addMethod: self inCategory: aCategory .
^ self
%

category: 'Reloading Decompiled Methods'
classmethod: CompiledMethod
_new

"Private.  Used to regenerate a method from decompiled form."

^ super new
%

category: 'Storing and Loading'
method: CompiledMethod
writeTo: aPassiveObject

"Instances of CompiledMethod cannot be converted to passive form.  This method
 writes nil to aPassiveObject and stops GemStone Smalltalk execution with a
 notifier."

aPassiveObject writeObject: nil.
self _error: #rtErrAttemptToPassivateInvalidObject.
%

category: 'Repository Conversion'
method: CompiledMethod
verifyContents

"Private.  Returns true if all of the objects referenced by the receiver exist,
 false otherwise.  For repository conversion testing."

<primitive: 473>
%

