!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: constrpatheval.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   ConstrainedPathEvaluator, PathEvaluator, Array, SequenceableCollection,
!   Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods ConstrainedPathEvaluator
removeallclassmethods ConstrainedPathEvaluator
protectmethods

category: 'For Documentation Installation only'
classmethod: ConstrainedPathEvaluator
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Instance methods for ConstrainedPathEvaluator

category: 'Query Select'
method: ConstrainedPathEvaluator
_findAllNonNilValuesWithComparison: selector to: aValue

""
<protected>

| obj endObj tmpHolder bag incomplete |
tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
    nil == (obj := bag _at: i)
        ifFalse: [
            endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

            ( nil ~~ endObj _and:
            [ (incomplete ~~ endObj) _and:
            [ endObj perform: selector with: aValue ] ] )
                ifTrue: [ tmpHolder add: obj ]
        ]
].
^ tmpHolder completeBag
%

category: 'Query Select'
method: ConstrainedPathEvaluator
findAllValuesGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

""
<protected>

| obj endObj tmpHolder bag incomplete |
" if query on NSC elements, execute optimized method (no traversal) "
self isIndexOnRootNsc
  ifTrue: [
    ^ super _findAllValuesGreaterThan: val1
      andEquals: bool1
      andLessThan: val2
      andEquals: bool2
  ].

" see if we can optimize the comparison operations "
( val1 ~~ nil _and: [ val2 ~~ nil _and: [ self _canUseSimpleComparisons ] ] )
  ifTrue: [
    " execute method that uses <, >, etc for comparisons "
    ^ self _findAllValuesGreaterThan: val1
      andEquals: bool1
      andLessThan: val2
      andEquals: bool2
  ].

tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.
    
              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.
    
              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.
    
              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ].

^ tmpHolder completeBag
%

category: 'Query Select'
method: ConstrainedPathEvaluator
_findAllValuesWithComparison: selector to: aValue

""
<protected>

| obj endObj tmpHolder bag incomplete |

tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
    nil == (obj := bag _at: i)
        ifFalse: [
            endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

            ( incomplete ~~ endObj _and:
            [ endObj perform: selector with: aValue ] )
                ifTrue: [ tmpHolder add: obj ]
        ]
].
^ tmpHolder completeBag
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
_findFirstNonNilValueWithComparison: selector to: aValue

""
<protected>

| obj endObj bag incomplete |
incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
    nil == (obj := bag _at: i)
        ifFalse: [
            endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

            ( nil ~~ endObj _and:
            [ (incomplete ~~ endObj) _and:
            [ endObj perform: selector with: aValue ] ] )
                ifTrue: [ ^ obj ]
        ]
].
^ #_incompletePathTraversal
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
findFirstValueGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

""
<protected>

| obj endObj bag incomplete |
incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete.

              ( incomplete ~~ endObj _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
  ].

^ incomplete
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
_findFirstValueWithComparison: selector to: aValue

""
<protected>

| incomplete bag obj endObj |

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
  nil == (obj := bag _at: i)
    ifFalse: [
      endObj := self _traverse: obj
        offsets: ivOffsets
        constraints: constraintClasses
        incomplete: incomplete.

      ( incomplete ~~ endObj _and: [ endObj perform: selector with: aValue ] )
        ifTrue: [ ^ obj ]
    ]
].

^ #_incompletePathTraversal
%

category: 'Traversing'
method: ConstrainedPathEvaluator
_initTraverseCache

""

<protected primitive: 185>
self _primitiveFailed: #_initTraverseCache .
self _uncontinuableError
%

category: 'Traversing'
method: ConstrainedPathEvaluator
_traverse: obj
offsets: anArray
constraints: constraints
incomplete: incompletePathTraversal

"Traverse the given object along the given instance variable offsets.  If a nil
 is encountered before the end of the path, returns incompletePathTraversal.
 If the primitive fails, recalculate the instance variable offsets based upon
 the class of obj, and try again."

<protected primitive: 219>

self _calculateIvOffsetsOnKind: obj class.
^ super _traverseObject: obj
%

category: 'Traversing'
method: ConstrainedPathEvaluator
_traverseObject: anObject

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, returns the
 incompletePathTraversalObject."
<protected>

^ self _traverse: anObject
  offsets: ivOffsets
  constraints: constraintClasses
  incomplete: #_incompletePathTraversal
%

category: 'Traversing'
method: ConstrainedPathEvaluator
_traverseObject: anObject
incompletesInto: incompleteArray
incomplete: incomplete

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, place the object
 into the appropriate Array in the incompleteArray of Arrays."
<protected>

| nextObj |
" if could not traverse to the end of the path , update the incompleteArray "
incomplete == 
(nextObj := self _traverse: anObject
  offsets: ivOffsets
  constraints: constraintClasses
  incomplete: incomplete)
  ifFalse: [ ^ nextObj ].

^ super _traverseObject: anObject
  incompletesInto: incompleteArray
  incomplete: incomplete
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
findFirstValueEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

| obj bag incomplete |
incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
  nil == (obj := bag _at: i)
    ifFalse: [
      (self _traverse: obj
         offsets: ivOffsets
         constraints: constraintClasses
         incomplete: incomplete) = aValue ifTrue: [ ^ obj ]
        ].
].

^ incomplete
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
findFirstValueGreaterThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

| obj endObj bag incomplete |
( QueryExecuter nilOrderDefinedFor: self lastElementClass )
  ifFalse: [
    ^ super findFirstValueGreaterThanKey: aValue andEquals: aBoolean
  ].

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
aBoolean
    ifTrue: [
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverse: obj
                        offsets: ivOffsets
                        constraints: constraintClasses
                        incomplete: incomplete.

                    ( nil ~~ endObj _and:
                    [ (incomplete ~~ endObj) _and:
                    [ endObj _idxCompareGreaterThanOrEqual: aValue ] ] )
                        ifTrue: [ ^ obj ]
                ]
        ]
    ]
    ifFalse: [
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverse: obj
                        offsets: ivOffsets
                        constraints: constraintClasses
                        incomplete: incomplete.

                    ( nil ~~ endObj _and:
                    [ (incomplete ~~ endObj) _and:
                    [ endObj _idxCompareGreaterThan: aValue ] ] )
                        ifTrue: [ ^ obj ]
                ]
        ]
    ].

^ incomplete
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
findFirstValueIdenticalTo: aValue

"Returns the first value that satisfies the query."
<protected>

| obj bag incomplete |
incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
    nil == (obj := bag _at: i)
        ifFalse: [
            aValue == 
            (self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: incomplete)
                ifTrue: [ ^ obj ]
        ]
].
^ incomplete
%

category: 'Query Detect'
method: ConstrainedPathEvaluator
findFirstValueLessThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

| obj endObj bag incomplete |
( QueryExecuter nilOrderDefinedFor: self lastElementClass ) 
  ifFalse: [
    ^ super findFirstValueLessThanKey: aValue andEquals: aBoolean
  ].

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
aBoolean
    ifTrue: [
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverse: obj
                        offsets: ivOffsets
                        constraints: constraintClasses
                        incomplete: incomplete.

                    ( nil ~~ endObj _and:
                    [ (incomplete ~~ endObj) _and:
                    [ endObj _idxCompareLessThanOrEqual: aValue ] ] )
                        ifTrue: [ ^ obj ]
                ]
        ]
    ]
    ifFalse: [
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverse: obj
                        offsets: ivOffsets
                        constraints: constraintClasses
                        incomplete: incomplete.

                    ( nil ~~ endObj _and:
                    [ (incomplete ~~ endObj) _and:
                    [ endObj _idxCompareLessThan: aValue ] ] )
                        ifTrue: [ ^ obj ]
                ]
        ]
    ].

^ incomplete
%

category: 'Query Select'
method: ConstrainedPathEvaluator
findAllValuesEqualTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

| obj tmpHolder bag incomplete |

tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
  nil == (obj := bag _at: i)
    ifFalse: [
      (self _traverse: obj
        offsets: ivOffsets
        constraints: constraintClasses
        incomplete: incomplete) = aValue
        ifTrue: [ tmpHolder add: obj ]
    ]
].
^ tmpHolder completeBag
%

category: 'Query Select'
method: ConstrainedPathEvaluator
findAllValuesIdenticalTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

| obj tmpHolder bag incomplete |

" optimize query for objects in the NSC itself "
(self at: 1) ==  #'' 
    ifTrue: [ ^ self _allIdenticalTo: aValue in: nsc ].

tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.
self _initTraverseCache.
1 to: bag size do: [ :i |
  nil == (obj := bag _at: i)
    ifFalse: [
      aValue ==
      (self _traverse: obj
        offsets: ivOffsets
        constraints: constraintClasses
        incomplete: incomplete)
        ifTrue: [ tmpHolder add: obj ]
    ]
].
^ tmpHolder completeBag
%

!----------------------------------------------------------------------
! Optimized comparison operations 
!----------------------------------------------------------------------

category: 'Query Support'
method: ConstrainedPathEvaluator
_canUseSimpleComparisons

"Returns whether the receiver can use simple comparison operations (<, >, etc)
 rather than indexing-specific ones (_idxCompareGreaterThan:).  This can
 be done if the last object along the path is the right class and the
 argument of the comparison is guaranteed not to be nil."
<protected>

| lastClass |
lastClass := self lastElementClass.

(lastClass isSubclassOf: Number)
  ifTrue: [
    " Floats need indexing-specific comparisons "
    ^ ( (lastClass isSubclassOf: BinaryFloat) _or: 
    [ lastClass isSubclassOf: DecimalFloat ] ) not
  ]
  ifFalse: [
    " Strings and Characters should behave correctly "
    ^ lastClass == String _or: [ lastClass == Character ]
  ]
%

category: 'Query Select'
method: ConstrainedPathEvaluator
_findAllValuesGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

"This method uses <, >, <=, >= for comparison operations, rather than
their _idx... equivalents."
<protected>

| obj endObj tmpHolder bag |
tmpHolder := NscBuilder
  for: nsc speciesForSelect new
  max: self sizeForNscBuilder.

bag := nsc _asIdentityBag.
self _initTraverseCache.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: nil.

              ( nil ~~ endObj _and:
              [ ( endObj >= val1 ) _and: [ endObj <= val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: nil.
    
              ( nil ~~ endObj _and:
              [ ( endObj >= val1 ) _and: [ endObj < val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: nil.
    
              ( nil ~~ endObj _and:
              [ ( endObj > val1 ) _and: [ endObj <= val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverse: obj
                offsets: ivOffsets
                constraints: constraintClasses
                incomplete: nil.
    
              ( nil ~~ endObj _and:
              [ ( endObj > val1 ) _and: [ endObj < val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ].

^ tmpHolder completeBag
%

category: 'Testing'
method: ConstrainedPathEvaluator
isPathConstrained

"Returns true if the path represented by the receiver is fully constrained."
<unprotected>

^ true
%
unprotectmethods
