!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: identityset.gs,v 1.7 2008-01-09 19:02:34 stever Exp $
!
! Superclass Hierarchy:
!   IdentitySet, IdentityBag, UnorderedCollection, Collection, Object.
!
!=========================================================================

removeallmethods IdentitySet
removeallclassmethods IdentitySet

category: 'For Documentation Installation only'
classmethod: IdentitySet
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'An IdentitySet is an IdentityBag in which any distinct object can occur only
 once.  Adding the same (identical) object to an IdentitySet multiple times is
 redundant.  The result is the same as adding it once.

 Since an IdentitySet is an identity-based collection, different (non-identical)
 but equivalent (equal) objects are treated as distinct from each other.  In
 Sets, they are not distinct.  Adding multiple equivalent objects to an
 IdentitySet yields an IdentitySet with as many elements as there are distinct
 equivalent objects.  In short, two different elements of an IdentitySet are
 never identical, but they may be equivalent.

 You can create subclasses of IdentitySet to restrict the kind of elements it
 contains.  When creating a subclass of IdentitySet, you must specify a class as
 the aConstraint argument.  This class is called the element kind of the new
 subclass.  For each instance of the new subclass, the class of each element
 must be of the element kind.' .
doc documentClassWith: txt.

self description: doc.
%

!==========================================================================
! Below this point: only method definitions in fileout order
!==========================================================================

! ------------------- Class methods for IdentitySet
! ------------------- Instance methods for IdentitySet
category: 'Adding'
method: IdentitySet
add: anObject

"Adds anObject to the receiver only if it is not already an element of the
 receiver.  Returns anObject.  Has no effect if anObject is nil."

"Note: In GemStone 4.1, this method returned the receiver."

<primitive: 303>
self _primitiveFailed: #add: .
self _uncontinuableError
%
category: 'Adding'
method: IdentitySet
add: anObject withOccurrences: anInteger

"Disallowed.  Each element of an IdentitySet must be unique."

self shouldNotImplement: #add:withOccurrences:
%
category: 'Adding'
method: IdentitySet
addValue: anObject

"Adds anObject to the receiver only if it is not already an element of the
 receiver, and if the receiver does not contain an equivalent object.  Has no
 effect if anObject is nil.  Returns anObject."

(self includesValue: anObject) ifFalse:[ self add: anObject ].
^ anObject
%
category: 'Converting'
method: IdentitySet
asIdentitySet

"(R) Returns an IdentitySet with the contents of the receiver."

^ self.
%
category: 'Converting'
method: IdentitySet
asSet

"(R) Disallowed.  IdentitySets cannot be converted to Sets because it could
 result in the loss of some of the constituent elements."

^ self shouldNotImplement: #asSet
%
category: 'Indexing Support'
method: IdentitySet
_isLastOccurrenceInIndexObjectsFor: anObject
  "Returns true if the given object is maintained in the indexing objects for one
   occurrence. For identitySet this method can be overridden to return true, 
   if there is no set-valued term based on the receiver."
((self _indexedPaths ~~ nil) _and: [self _indexedPaths _hasASetValuedTerm])
  ifTrue: [^super _isLastOccurrenceInIndexObjectsFor: anObject].
^true
%
