!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: largepo.gs,v 1.7 2008-01-09 19:02:34 stever Exp $
!
! Superclass Hierarchy:
!   LargePositiveInteger, Integer, Number, Magnitude, Object.
!
!=========================================================================

removeallmethods LargePositiveInteger
removeallclassmethods LargePositiveInteger

category: 'For Documentation Installation only'
classmethod: LargePositiveInteger
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'Instances of LargePositiveInteger represent positive integers whose values are
 greater than the maximum SmallInteger (2 to the 29th power minus 1).  Each
 instance of LargePositiveInteger is stored as an Array of bytes, where each
 double-byte pair represents a base 32768 digit.  The first two bytes in that
 Array constitute the least significant base 32768 digit, and the last two 
 bytes are the most significant base 32768 digit.  Within each digit, the
 least significant byte is first, followed by the more significant byte.

 Coercion between LargePositiveInteger and SmallInteger occurs automatically.'.
doc documentClassWith: txt.

self description: doc.
%

category: 'Accessing'
method: LargePositiveInteger
at: anIndex

"Disallowed."

self shouldNotImplement: #at:
%

category: 'Accessing'
method: LargePositiveInteger
at: anIndex put: aNumber

"Disallowed.  You may not change the value of a LargePositiveInteger."

self shouldNotImplement: #at:put:
%

category: 'Testing'
method: LargePositiveInteger
negative

"Returns false."

^false
%

category: 'Testing'
method: LargePositiveInteger
positive

"Returns true."

^true
%

category: 'Truncation and Rounding'
method: LargePositiveInteger
truncated

"Returns the receiver."

| size partial |

size := self _digitLength.
(size == 0)
   ifTrue: [^0].
partial := self _digitAt: size.
(partial == 0)
   ifTrue: [^(self _growby: -1) truncated].
(size <= 2 _and:[ (self _digitAt: 2) < 16384 ])
   ifTrue: "Convert back to a SmallInteger."
      [partial := self _digitAt: 1.
       partial := partial + ((self _digitAt: 2) bitShift: 15).
       ^partial]
%

category: 'Arithmetic'
method: LargePositiveInteger
abs

"Returns a LargePositiveInteger that is the absolute value of the receiver."

^ self
%

category: 'Arithmetic'
method: LargePositiveInteger
negated

"Returns a LargeNegativeInteger that is the negation of the receiver."

| x size n byte |

size := self _digitLength.
x := Integer _new: size neg: true.
n := 1.
[n <= size]
   whileTrue: [
      byte := self _digitAt: n.
      x _digitAt: n put: byte.
      n := n + 1].
^ x truncated
%

category: 'Bit Manipulation'
method: LargePositiveInteger
highBit

"Returns the index of the high-order bit that is set in the binary
 representation of the receiver."

^self _lastDigit highBit + (15 * (self _digitLength - 1))
%

