!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: rangeindex.gs,v 1.7 2008-01-09 19:02:35 stever Exp $
!
! Superclass Hierarchy:
!   RangeEqualityIndex, IdentityIndex, Array, SequenceableCollection,
!   Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods RangeEqualityIndex
removeallclassmethods RangeEqualityIndex
protectmethods

category: 'For Documentation Installation only'
classmethod: RangeEqualityIndex
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Class methods for RangeEqualityIndex
category: 'Constants'
classmethod: RangeEqualityIndex
isBasicClass: aClass

"Returns whether the given class should use BtreeBasicLeafNodes to store
 B-tree mappings."
<protected>

aClass == nil
    ifTrue: [ ^ false ].

aClass isSpecial
    ifTrue: [ ^ true ].

#[ String, 
   DoubleByteString,   
           "EUCString no longer supported by encryption"
   Number, "all kernel subclasses of Number" 
   Date,
"   DateTime,"
   Time, 
   AbstractCharacter "Character and JISCharacter handled here" 
  ] do: [ :basicClass |
        (aClass == basicClass _or: [ aClass inheritsFrom: basicClass ] )
            ifTrue: [ ^ true ]
        ].

^ false
%

category: 'Constants'
classmethod: RangeEqualityIndex
_maxCharsForStringComparison

"Returns the maximum number of Characters in a String used to differentiate in
 comparison operations (if Strings differ after this number of Characters, they
 are considered equal)."

<protected primitive: 161>
self _primitiveFailed: #_maxCharsForStringComparison .
self _uncontinuableError
%

category: 'Instance Creation'
classmethod: RangeEqualityIndex
newWithLastElementClass: aClass

"Create a new instance and initialize its B-tree root."
<protected>

| newOne |
newOne := super new.
newOne lastElementClass: aClass.
newOne btreeRoot: aClass btreeLeafNodeClass new.
^ newOne
%

category: 'Accessing'
method: RangeEqualityIndex
lastElementClassVersion: versionNum

"Returns the value of the instance variable 'lastElementClass'."

<protected>

(versionNum = 50)
  ifTrue: [ ^ lastElementClass ].

(versionNum = 41)
  ifTrue: [ ^ progress ].

self halt: 'invalid version num specified'.
%


! ------------------- Instance methods for RangeEqualityIndex
category: 'Error Handling'
method: RangeEqualityIndex
_errorObjectNotInBtree: key value: value

"An entry for the key/value was not present in the B-tree."
<protected>

^ self _error: #rtErrRangeEqualityIndexObjectNotInBtree args: #[ key, value ]
%

category: 'Searching'
method: RangeEqualityIndex
_findAllValuesGreaterThan: val1 andEquals: bool1
andLessThan: val2 andEquals: bool2

"Returns a RangeIndexReadStream that iterates over all objects in the index
 that satisfy the query."
<protected>

| array1 array2 stream doNext |
( (self _canCompareWith: val1) _and: [ self _canCompareWith: val2 ] )
    ifFalse: [ self _error: #assocErrBadComparison ].

stream := self readStreamClass new rangeIndex: self.
" check to see if the query is even satisfiable "
( (val1 _idxCompareGreaterThan: val2) _or:
[ val1 = val2 _and: [ (bool1 & bool2) not ] ] )
    ifTrue: [
        stream currentStack: (Array with: 0).
        ^ stream
    ].

doNext := false.
array1 := Array new.
" see if querying > or >= "
bool1
    ifTrue: [ " if >=, perform < and do an '_advance' operation later "
        (btreeRoot _findAllValuesLessThanKey: val1 andEquals: false into: array1)
            ifTrue: [ doNext := true ]
            ifFalse: [ " all were greater than "
                btreeRoot _putFirstIndexOfFirstChildInto: array1
            ]
    ]
    ifFalse: [ " if >, ask the B-tree specifically for > "
        (btreeRoot _findAllValuesGreaterThanKey: val1 into: array1)
            ifFalse: [ " none were greater than "
                stream currentStack: (Array with: 0).
                ^ stream
            ]
    ].

array2 := Array new.
" ask the B-tree for the second boundary of the query result "
(btreeRoot _findAllValuesLessThanKey: val2 andEquals: bool2 into: array2)
    ifFalse: [ " none were found less than "
        stream currentStack: (Array with: 0).
        ^ stream
    ].

stream currentStack: array1.
stream endIndex: (array2 at: 1).
stream endNode: (array2 at: 2).

(stream endNode == (array1 at: 2) _and:
[ stream endIndex < (array1 at: 1) ] )
    ifTrue: [
        array1 at: 1 put: 0.
        ^ stream
    ].

doNext
    ifTrue: [ " advance the B-tree positioning "
        stream _advance.
    ].

^ stream
%

category: 'Searching'
method: RangeEqualityIndex
_findAllValuesGreaterThanKey: aKey andEquals: aBoolean

"Returns a RangeIndexReadStream that iterates over all objects in the index
 that satisfy the query."
<protected>

| array1 array2 stream doNext |
(self _canCompareWith: aKey)
    ifFalse: [ self _error: #assocErrBadComparison ].

stream := self readStreamClass new rangeIndex: self.

doNext := false.
array1 := Array new.
" see if querying > or >= "
aBoolean
    ifTrue: [ " if >=, perform < and do a 'next' operation later "
        (btreeRoot _findAllValuesLessThanKey: aKey andEquals: false into: array1)
            ifTrue: [ doNext := true ]
            ifFalse: [ " all were greater than "
                btreeRoot _putFirstIndexOfFirstChildInto: array1
            ]
    ]
    ifFalse: [ " if >, ask the B-tree specifically for > "
        (btreeRoot _findAllValuesGreaterThanKey: aKey into: array1)
            ifFalse: [ " none were greater than "
                stream currentStack: (Array with: 0).
                ^ stream
            ]
    ].

array2 := Array new.
btreeRoot _putLastIndexOfLastChildInto: array2.

stream currentStack: array1.
stream endIndex: (array2 at: 1).
stream endNode: (array2 at: 2).

doNext ifTrue: [ stream _advance ].

^ stream
%

category: 'Searching'
method: RangeEqualityIndex
_findAllValuesLessThanKey: aKey andEquals: aBoolean

"Returns a RangeIndexReadStream that iterates over all objects in the index
 that satisfy the query."
<protected>

| array1 array2 stream |
(self _canCompareWith: aKey)
    ifFalse: [ self _error: #assocErrBadComparison ].

stream := self readStreamClass new rangeIndex: self.
array2 := Array new.
" ask the B-tree to make the search and put the result path into array2 "
(btreeRoot _findAllValuesLessThanKey: aKey andEquals: aBoolean into: array2)
    ifTrue: [
        array1 := Array new.
        " put path to first entry of the B-tree into array1 "
        btreeRoot _putFirstIndexOfFirstChildInto: array1.
        stream currentStack: array1.
        stream endIndex: (array2 at: 1).
        stream endNode: (array2 at: 2)
    ]
    ifFalse: [ " none were found "
        stream currentStack: (Array with: 0)
    ].

^ stream
%

category: 'Removing'
method: RangeEqualityIndex
_removeBtreeEntriesForKey: aKey

"Removes all entries in the B-tree that have a key identical to aKey.
 Returns an Array of values corresponding to each entry that was removed."
<protected>

| stream vals |
" first we need to find all values that have aKey as the key "
stream := self _findAllValuesGreaterThanKey: aKey andEquals: true.
vals := Array new.

[ stream _btreeAtEnd not _and:
[ stream _peekKey _idxCompareEqualTo: aKey ] ] whileTrue: [
    aKey == stream _peekKey
        ifTrue: [ vals add: stream _peekValue ].
    stream _btreeNext
].
" now remove the entry for each value "
1 to: vals size do: [ :i |
  (self btreeRemoveKey: aKey value: (vals at: i))
     "ifFalse: [ self _errorObjectNotInBtree: aKey value: (vals at: i) ]"
].

^ vals
%

category: 'Updating'
method: RangeEqualityIndex
btreeAt: aKey put: aValue

"Insert the key/value pair into the root B-tree node.  Must check to see if a
 split occurred.  If so, create a new root node and put the old root and the
 new split node into the new root."
<protected>

| returnNode node |
(self _canCompareWith: aKey)
    ifFalse: [
      ^ self _error: #rtErrRangeEqualityIndexInvalidClassKindForBtree
        args: #[ aKey, lastElementClass ]
    ].

returnNode := btreeRoot btreeAt: aKey put: aValue.

" see if a split occurred "
returnNode == btreeRoot
    ifFalse: [ " returnNode is the second half of the split "
        " create the new parent node "
        node := btreeRoot parentNodeClass new.
        node lastValue: returnNode lastValue.

        " insert the first half (the original root) "
        node
            _insertKey: (btreeRoot _at: btreeRoot _lastKeyIndex)
            value: btreeRoot
            atIndex: 1.
        " insert the second half "
        node
            _insertKey: (returnNode _at: returnNode _lastKeyIndex)
            value: returnNode
            atIndex: (node entrySize + 1).

        btreeRoot := node.
    ]
%

category: 'Updating'
method: RangeEqualityIndex
btreeRemoveKey: aKey value: aValue

"Removes the key and value from the B-tree root node.  Must check to see if
 a merge occurred in the root such that the root only contains a single
 entry.  If so, make the single entry the new root.  Returns whether the
 removal occurred."
<protected>

(btreeRoot removeKey: aKey value: aValue)
    ifTrue: [
        ((btreeRoot numElements == 1) _and: [ btreeRoot isLeaf not ] )
            ifTrue: [ btreeRoot := btreeRoot at: (btreeRoot _lastKeyIndex - 1) ].
        ^ true
    ]
    ifFalse: [ ^ false ].
%

category: 'Accessing'
method: RangeEqualityIndex
btreeRoot

"Returns the value of the instance variable 'btreeRoot'."
<protected>

^btreeRoot
%

category: 'Updating'
method: RangeEqualityIndex
btreeRoot: newValue

"Modify the value of the instance variable 'btreeRoot'."
<protected>

btreeRoot := newValue
%

category: 'Updating Indexes'
method: RangeEqualityIndex
changingSizeOfByteObject: aByteObject to: newSize

"The byte object's size is being changed.  For equality indexes, must update
 the B-tree."
<protected>

| vals newKey oldSize max |
oldSize := aByteObject size.

" If change is beyond the limit for comparisons, no change in B-tree position "
max := self class _maxCharsForStringComparison.
( oldSize > max _and: [ newSize > max ] )
    ifTrue: [ ^ self ].

vals := self _removeBtreeEntriesForKey: aByteObject.

" construct a pretend key "
newKey := aByteObject class new: newSize.
( newSize > 0 _and: [ aByteObject size > 0 ] )
    ifTrue: [
        aByteObject
            copyFrom: 1
            to: (oldSize min: newSize)
            into: newKey
            startingAt: 1
    ].

" for each value, add a new entry to the B-tree using the pretend key "
1 to: vals size do: [ :i | self btreeAt: newKey put: (vals at: i) ].

" now replace the pretend key with the real one "
btreeRoot _replaceKey: newKey with: aByteObject
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesEqualTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

(aValue class == Symbol _or:
[ aValue class == String ] )
  ifTrue: [
    ^ (self
      _findAllValuesGreaterThan: aValue
      andEquals: true
      andLessThan: aValue
      andEquals: true) makeNscFilterSymbols: (aValue class == Symbol)
  ].

^ (self 
    _findAllValuesGreaterThan: aValue 
    andEquals: true
    andLessThan: aValue
    andEquals: true) makeNsc
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

"Returns a set of all values that satisfy the query."
<protected>

^ (self
    _findAllValuesGreaterThan: val1
    andEquals: bool1
    andLessThan: val2
    andEquals: bool2) makeNsc
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesGreaterThanKey: aValue andEquals: aBoolean

"Returns a set of all values that satisfy the query."
<protected>

^ (self 
    _findAllValuesGreaterThanKey: aValue 
    andEquals: aBoolean) makeNsc
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesIdenticalTo: aValue

"Returns a set of all values that satisfy the query.  Since this is an identity
 comparison on a range index (which supports equality comparison), this
 method creates a B-tree read stream for equal values, then iterates through the
 values making the identity comparison."
<protected>

| array1 array2 tmpHolder stream |

" optimize query for objects in the NSC itself "
self isIndexOnRootNsc
    ifTrue: [ ^ self _allIdenticalTo: aValue in: nscRoot ].

" if can't make the comparison, then none are identical "
(self _canCompareWith: aValue)
    ifFalse: [ ^ nscRoot speciesForSelect new ].

" get boundary entries in the B-tree "
array1 := Array new.
(btreeRoot _findAllValuesLessThanKey: aValue andEquals: false into: array1)
    ifFalse: [ " all were greater than "
        btreeRoot _putFirstIndexOfFirstChildInto: array1
    ].

array2 := Array new.
(btreeRoot _findAllValuesLessThanKey: aValue andEquals: true into: array2)
    ifFalse: [ " none were found less than "
        ^ nscRoot speciesForSelect new
    ].

" create the read stream "
stream := RangeIndexReadStream new rangeIndex: self.
stream currentStack: array1.
stream endIndex: (array2 at: 1).
stream endNode: (array2 at: 2).

tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.
self size == 1
    ifTrue: [ stream _valuesAddInto: tmpHolder identicalTo: aValue ]
    ifFalse: [ stream _uniqueValuesAddInto: tmpHolder identicalTo: aValue ].
^ tmpHolder completeBag
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesLessThanKey: aValue andEquals: aBoolean

"Returns a set of all values that satisfy the query."
<protected>

" if <= nil "
( aValue == nil _and: [ aBoolean ] )
  ifTrue: [
    ^(self
      _findAllValuesGreaterThan: nil
      andEquals: true
      andLessThan: nil
      andEquals: true) makeNsc
  ]
  ifFalse: [
    ^ (self
      _findAllValuesGreaterThan: nil
      andEquals: (RangeEqualityIndex isBasicClass: lastElementClass) not
      andLessThan: aValue
      andEquals: aBoolean) makeNsc
  ]
%

category: 'Query Select'
method: RangeEqualityIndex
findAllValuesNotEqualTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

| preStream postStream tmpHolder |

" empirical analysis has shown that PathEvaluators are faster for this "
self size == 1
    ifFalse: [
        (self _canCompareWith: aValue)
            ifFalse: [ self _error: #assocErrBadComparison ].

      ^ self asPathEvaluator findAllValuesNotEqualTo: aValue
    ].

preStream := self _findAllValuesLessThanKey: aValue andEquals: false.
postStream := self _findAllValuesGreaterThanKey: aValue andEquals: false.

tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.
self size == 1
    ifTrue: [
        preStream _valuesAddInto: tmpHolder.
        postStream _valuesAddInto: tmpHolder.
    ]
    ifFalse: [
        preStream _uniqueValuesAddInto: tmpHolder.
        postStream _uniqueValuesAddInto: tmpHolder.
    ].
^ tmpHolder completeBag
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

| val |
(self _canCompareWith: aValue)
    ifFalse: [ self _error: #assocErrBadComparison ].

aValue class == Symbol
  ifTrue: [ val := btreeRoot _findFirstValueForKeyNoStrings: aValue ]
  ifFalse: [
    aValue class == String
      ifTrue: [ val := btreeRoot _findFirstValueForKeyNoSymbols: aValue ]
      ifFalse: [ val := btreeRoot _findFirstValueForKey: aValue ]
  ].

(val == #_incompletePathTraversal)
    ifTrue: [ ^ #_incompletePathTraversal ].

^ self _findFirstValueForKey: val
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueGreaterThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

| stream |
stream := self _findAllValuesGreaterThanKey: aValue andEquals: aBoolean.

stream _atEnd
    ifTrue: [ ^ #_incompletePathTraversal ]
    ifFalse: [ ^ stream getNext ]
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

"Returns the first value that satisfies the query."
<protected>

| stream |
stream := self
    _findAllValuesGreaterThan: val1
    andEquals: bool1
    andLessThan: val2
    andEquals: bool2.

stream _atEnd
    ifTrue: [ ^ #_incompletePathTraversal ]
    ifFalse: [ ^ stream getNext ]
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueIdenticalTo: aValue

"Returns the first value that satisfies the query.  Since this is an identity
 comparison on a range index (which supports equality comparison), create a
 BtreeReadStream for equal values, then iterate through the values making
 the identity comparison."
<protected>

| array1 array2 stream |
" if can't make the comparison, then none are identical "
(self _canCompareWith: aValue)
    ifFalse: [ ^ #_incompletePathTraversal ].

" get boundary entries in the B-tree "
array1 := Array new.
(btreeRoot _findAllValuesLessThanKey: aValue andEquals: false into: array1)
    ifFalse: [ " all were greater than or equal to "
        btreeRoot _putFirstIndexOfFirstChildInto: array1
    ].

array2 := Array new.
(btreeRoot _findAllValuesLessThanKey: aValue andEquals: true into: array2)
    ifFalse: [ " none were found less than or equal to "
        ^ #_incompletePathTraversal
    ].

" create the read stream "
stream := BtreeReadStream new.
stream currentStack: array1.
stream endIndex: (array2 at: 1).
stream endNode: (array2 at: 2).

" iterate through the stream, checking for an identical value "
[ stream atEnd not ] whileTrue: [
    (stream _peekKey == aValue)
        ifTrue: [ ^ self _findFirstValueForKey: stream _peekValue ].
    stream _btreeNext
].
^ #_incompletePathTraversal
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueLessThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

| stream |
" if <= nil "
( aValue == nil _and: [ aBoolean ] )
  ifTrue: [
    stream := self 
      _findAllValuesGreaterThan: nil
      andEquals: true
      andLessThan: nil
      andEquals: true.
  ]
  ifFalse: [
    stream := self 
      _findAllValuesGreaterThan: nil
      andEquals: (RangeEqualityIndex isBasicClass: lastElementClass) not
      andLessThan: aValue
      andEquals: aBoolean.
  ].

stream _atEnd
    ifTrue: [ ^ #_incompletePathTraversal ]
    ifFalse: [ ^ stream getNext ]
%

category: 'Query Detect'
method: RangeEqualityIndex
findFirstValueNotEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

| stream totalStream obj bag incomplete |
" find the ones that are equal "
stream := self
  _findAllValuesGreaterThan: aValue
  andEquals: true
  andLessThan: aValue
  andEquals: true.

" if none are equal, pick the first one "
stream _atEnd
  ifTrue: [
    self hasSetValuedTerm
      ifTrue: [
        bag := nscRoot _asIdentityBag.
        1 to: bag size do: [ :i |
          nil ~~ (bag _at: i)
            ifTrue: [ ^ bag _at: i ]
        ].
        ^ #_incompletePathTraversal
      ]
      ifFalse: [
          incomplete := #_incompletePathTraversal.
          bag := nscRoot _asIdentityBag.
          1 to: bag size do: [ :i |
            obj := bag _at: i.
            ((self traverseObject: obj) == incomplete)
              ifFalse: [ ^ obj ]
          ].
          ^ #_incompletePathTraversal
      ]
  ].

" get a stream over the entire NSC "
totalStream := RangeIndexReadStream on: self.
" see if very first object is not equal "
totalStream _peekKey ~= stream _peekKey
  ifTrue: [ ^ totalStream getNext ].

" see if end of original stream is the very last object "
(totalStream _hasSameEndAs: stream )
  ifTrue: [ ^ #_incompletePathTraversal ]
  ifFalse: [
    " advance the stream to the last object "
    totalStream _positionToEnd.
    ^ totalStream getNext
  ]
%

category: 'Testing'
method: RangeEqualityIndex
isIdentityIndex

"Returns false."
<unprotected>

^ false
%

category: 'Testing'
method: RangeEqualityIndex
isRangeEqualityIndex

"Returns true."
<unprotected>

^ true
%

category: 'Accessing'
method: RangeEqualityIndex
lastElementClass

"Returns the value of the instance variable 'lastElementClass'."
<protected>

^lastElementClass
%

category: 'Repository Conversion'
method: RangeEqualityIndex
lastElementClass41

"This method is a kludge.  The right way would be to have corresponding
 obsolete classes because instance variables were added going from 4.1 to 5.0."

<protected>

^progress
%

category: 'Repository Conversion'
method: RangeEqualityIndex
pathComponentsString41

"This method is a kludge.  The right way would be to have corresponding
 obsolete classes because instance variables were added going from 4.1 to 5.0."

<protected>

| objSize newRangeIndex numInstVars |

newRangeIndex := RangeEqualityIndex new.

numInstVars := RangeEqualityIndex allInstVarNames size.
objSize := self _primitiveSize.

(objSize > numInstVars)
  ifTrue: [
    6 to: numInstVars do: [ :i |
      newRangeIndex add: (self instVarAt: i).
      ].

    1 to: self size do: [ :i |
      newRangeIndex add: (self at: i).
      ].
    ]
  ifFalse: [
    6 to: objSize do:  [ :i |
      newRangeIndex add: (self instVarAt: i).
      ].
    ].

^ newRangeIndex pathComponentsString.
%

category: 'Repository Conversion'
method: RangeEqualityIndex
pathComponentsString41

"This method is a kludge.  The right way would be to have corresponding
 obsolete classes because instance variables were added going from 4.1 to 5.0."

<protected>

| objSize newRangeIndex numInstVars |

newRangeIndex := RangeEqualityIndex new.

numInstVars := RangeEqualityIndex allInstVarNames size.
objSize := self _primitiveSize.

(objSize > numInstVars)
  ifTrue: [
    6 to: numInstVars do: [ :i |
      newRangeIndex add: (self instVarAt: i).
      ].

    1 to: self size do: [ :i |
      newRangeIndex add: (self at: i).
      ].
    ]
  ifFalse: [
    6 to: objSize do:  [ :i |
      newRangeIndex add: (self instVarAt: i).
      ].
    ].

^ newRangeIndex pathComponentsString.
%

category: 'Updating'
method: RangeEqualityIndex
lastElementClass: newValue

"Modify the value of the instance variable 'lastElementClass'."
<protected>

lastElementClass := newValue
%

category: 'Accessing'
method: RangeEqualityIndex
lastPathComponentsDictionaryOffset

"Returns the last index into the path components list that is mapped in the
 index dictionary (as opposed to the B-tree)."
<protected>

^ self size - 1
%

category: 'Accessing'
method: RangeEqualityIndex
readStreamClass

"Returns the class of read stream to create for query results."
<protected>

^ RangeIndexReadStream
%

category: 'Updating Indexes'
method: RangeEqualityIndex
updatingByteObject: aByteObject startingAt: startPt withNewValue: newValue

"The byte object is being updated.  For equality indexes, must update the
 B-tree."
<protected>

| vals newKey |
" If change is beyond the limit for comparisons, no change in B-tree position "
(startPt > self class _maxCharsForStringComparison)
    ifTrue: [ ^ self ].

vals := self _removeBtreeEntriesForKey: aByteObject.

" construct a pretend key "
aByteObject class == Symbol
  ifTrue: [ newKey := String withAll: aByteObject ]
  ifFalse: [ newKey := aByteObject copy ].

newValue class == Character
  ifTrue: [
    newKey at: startPt put: newValue
  ]
  ifFalse: [
    newValue copyFrom: 1 to: newValue size into: newKey startingAt: startPt
  ].

" For each value, add a new entry to the B-tree using the pretend key "
1 to: vals size do: [ :i | self btreeAt: newKey put: (vals at: i) ].

" now replace the pretend key with the real one "
btreeRoot _replaceKey: newKey with: aByteObject
%

category: 'Accessing'
method: RangeEqualityIndex
changeToSegment: aSegment

"Assign the receiver and its B-tree to the given segment."
<protected>

self assignToSegment: aSegment.
btreeRoot changeToSegment: aSegment
%

category: 'Updating Indexes'
method: RangeEqualityIndex
addDirectMappingsFor: aBag
indexList: iList
updateIndexListSegments: aBoolean
commitInterval: anInteger

"Add an entry for an index on the elements of the NSC."
<protected>

| anObject firstPathTerm |
firstPathTerm := self firstPathTerm.
anInteger ~~ nil
  ifTrue: [ nscRoot _lockForIndexCreation ].

self lastPathTerm needsDepList
  ifTrue: [
    aBoolean
      ifTrue: [
        1 to: aBag size do: [ :i |
          anInteger ~~ nil
            ifTrue: [
              nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
            ].

          " Add an entry for an index on the elements of the NSC. "
          nil ~~ (anObject := aBag _at: i)
            ifTrue: [ " add an entry in the dependency list "
              anObject
                getDepListAndAddPathTerm: firstPathTerm
                withIVOffset: 1
                logging: false.
              iList addSegment: anObject segment logging: false.
            ].
          self btreeAt: anObject put: anObject.
        ]
      ]
      ifFalse: [
        1 to: aBag size do: [ :i |
          anInteger ~~ nil
            ifTrue: [
              nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
            ].

          " Add an entry for an index on the elements of the NSC. "
          nil ~~ (anObject := aBag _at: i)
            ifTrue: [ " add an entry in the dependency list "
              anObject
                getDepListAndAddPathTerm: firstPathTerm
                withIVOffset: 1
                logging: false
            ].
          self btreeAt: anObject put: anObject.
        ]
      ]
  ]
  ifFalse: [
    aBoolean
      ifTrue: [
        1 to: aBag size do: [ :i |
          anInteger ~~ nil
            ifTrue: [
              nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
            ].

          anObject := aBag _at: i.
          iList addSegment: anObject segment logging: false.
          self btreeAt: anObject put: anObject.
        ]
      ]
      ifFalse: [
        1 to: aBag size do: [ :i |
          anInteger ~~ nil
            ifTrue: [
              nscRoot _commitIndexCreation: self at: i commitInterval: anInteger
            ].

          anObject := aBag _at: i.
          self btreeAt: anObject put: anObject.
        ]
      ]
  ].
%

category: 'Updating Indexes'
method: RangeEqualityIndex
preIndexCreation

"Determine if a merge sort is necessary.  If so, place a path sorter
 as the btreeRoot (will be replaced with a B-tree node after the merge
 sort).  Returns the original btreeRoot node. "
<protected>

| saveNode |
super preIndexCreation.

" check if all mappings will fit in a single B-tree node "
nscRoot size > btreeRoot class maxNumberOfElements
  ifTrue: [
    saveNode := btreeRoot.
    btreeRoot := PathSorter on: #[ self ] directions: #( true ).
    " initialize offset of current node to update in sort node Array "
    btreeRoot lastNodeOffset: 1.
    ^ saveNode
  ].
^ btreeRoot
%

category: 'Updating Indexes'
method: RangeEqualityIndex
postIndexCreation: originalBtreeRoot commitInterval: anInteger

"Use a merge sort to create a complete B-tree from a PathSorter
 stored in the btreeRoot instance variable."
<protected>

" see if merge sort necessary "
btreeRoot isBtreeNode
  ifFalse: [ | leafNodes |
    " create an Array containing an empty B-tree node "
    leafNodes := #[ originalBtreeRoot ].
    " produce an Array containing many B-tree leaf nodes, all sorted "
    btreeRoot sortIntoBtreeNodes: leafNodes commitInterval: anInteger.
    " now build up interior nodes for the leaves "
    btreeRoot := btreeRoot createInteriorNodesFor: leafNodes.
  ].

super postIndexCreation: originalBtreeRoot commitInterval: anInteger.
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_count

"Returns the number of entries in the leaf nodes of the B-tree.  This count
 includes objects whose last path object is nil."
<protected>

| count |
self checkAuthorization.
count := 0.
btreeRoot _preOrderDo: [ :node |
    node isLeaf
        ifTrue: [ count := count + node numElements ]
].
^ count
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_max

"Returns the object with the maximum value for the last object in the path."
<protected>

| stream |
self checkAuthorization.
stream := RangeIndexReadStream on: self.
stream _positionToEnd.
stream _atEnd
    ifTrue: [ ^ nil ]
    ifFalse: [ ^ stream getNext ]
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_mean

"Returns the mean of non-nil, non-NaNs of the values of the last object (a
 Number) in the path."
<protected>

| count sum eSize |
self checkAuthorization.
(lastElementClass isSubclassOf: Number)
    ifFalse: [ ^ self _error: #rtErrRangeEqualityIndexNotANumber ].

count := 0.
sum := 0.
eSize := btreeRoot entrySize.
btreeRoot _preOrderDo: [ :node |
    node isLeaf
        ifTrue: [ | key |
            2 to: (node numElements * eSize) by: eSize do: [ :i |
                ( nil ~~ (key := node _at: i) _and: [ key _isNaN not ] )
                    ifTrue: [
                        sum := sum + key.
                        count := count + 1
                    ]
            ]
        ]
].
^ sum asFloat / count
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_median

"Returns the median of non-nil, non-NaNs of the values of the last object (a
 Number) in the path."
<protected>

| array1 array2 node count numToMidPt num val1 val2 |
self checkAuthorization.
(lastElementClass isSubclassOf: Number)
    ifFalse: [ ^ self _error: #rtErrRangeEqualityIndexNotANumber ].

array1 := Array new.
array2 := Array new.
btreeRoot _putLeavesIn: array1 numNonNilsIn: array2.

count := 0.
1 to: array2 size do: [ :i |
    count := count + (array2 at: i)
].

numToMidPt := count // 2.

1 to: array1 size do: [ :i |
    (array2 at: i) < numToMidPt
        ifTrue: [ numToMidPt := numToMidPt - (array2 at: i) ]
        ifFalse: [
            node := array1 at: i.
            num := numToMidPt + node numElements - (array2 at: i).
            count even
                ifTrue: [
                    val1 := node _at: (num - 1) * node entrySize + 2.
                    val2 := node _at: num * node entrySize + 2.
                    ^ (val1 + val2) asFloat / 2
                ]
                ifFalse: [
                    ^ node _at: num * node entrySize + 2
                ]
        ]
]
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_min

"Returns the object with the minimum value for the last object in the path."
<protected>

self checkAuthorization.
^ self findFirstValueGreaterThanKey: nil
    andEquals: false
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_nonNilCount

"Returns the number of entries in the leaf nodes of the B-tree.  This count
 does not include objects whose last path object is nil."
<protected>

| count |
self checkAuthorization.
count := 0.
btreeRoot _preOrderDo: [ :node |
    node isLeaf
        ifTrue: [ count := count + node _numNonNilElements ]
].
^ count
%

category: 'Query Aggregate'
method: RangeEqualityIndex
_sum

"Returns the sum of non-nil, non-NaNs of the values of the last object (a
 Number) in the path."
<protected>

| sum eSize |
self checkAuthorization.
(lastElementClass isSubclassOf: Number)
    ifFalse: [ ^ self _error: #rtErrRangeEqualityIndexNotANumber ].

sum := 0.
eSize := btreeRoot entrySize.
btreeRoot _preOrderDo: [ :node |
    node isLeaf
        ifTrue: [ | key |
            2 to: (node numElements * eSize) by: eSize do: [ :i |
                ( nil ~~ (key := node _at: i) _and: [ key _isNaN not ] )
                    ifTrue: [ sum := sum + key ]
            ]
        ]
].
^ sum
%

category: 'Testing'
method: RangeEqualityIndex
_canCompareWith: aKey

"Returns whether the receiver can make B-tree comparisons with the given key."
<protected>

aKey == nil
    ifTrue: [ ^ true ].

^ btreeRoot _canCompare: aKey withClass: lastElementClass
%

category: 'Query Detect'
method: RangeEqualityIndex
_findFirstKeyNotIdenticalTo: aKey atTerm: pathTerm

"Returns the first key in the receiver's B-tree that is not identical to the
 given key."
<protected>

| stream key |
stream := BtreeReadStream on: btreeRoot.
[ stream atEnd ] whileFalse: [
  key := stream _peekKey.
  (key == aKey)
    ifFalse: [ ^ key ].
  stream _btreeNext
].
^ #_incompletePathTraversal
%

category: 'Updating'
method: RangeEqualityIndex
_setPathTermState

"For each path term but the last, indicate the need to update the index
 dictionary.  For the last path term, indicate the need to update the B-tree.
 Indicate if the last object along the path needs a dependency list."
<protected>

| lastPathTerm |
1 to: self size - 1 do: [ :i |
    (self at: i) updateDict: indexDictionary.
].
lastPathTerm := self at: self size.
lastPathTerm updateBtree: self.

" strings or binary floats need a dependency list "
( (lastElementClass inheritsFrom: CharacterCollection) _or:
[ lastElementClass inheritsFrom: BinaryFloat ] )
    ifTrue: [
        lastPathTerm needsDepList: true.
        ^ self
    ].
%
category: 'Sorting'
method: RangeEqualityIndex
_getSecondarySorterOn: indexObjs directions: theBooleans

"Returns a PathSorter if one is needed to perform secondary sorting;
 otherwise returns nil."
<protected>

| directions traversers |

indexObjs size > 1
  ifTrue: [
    directions := Array new: indexObjs size - 1.
    theBooleans copyFrom: 2 to: indexObjs size into: directions startingAt: 1.
    traversers := Array new: indexObjs size - 1.
    indexObjs copyFrom: 2 to: indexObjs size into: traversers startingAt: 1.
    ^ PathSorter on: traversers directions: directions.
  ]
  ifFalse: [ ^ nil ].
%

category: 'Sorting'
method: RangeEqualityIndex
_getObjectsWithNilOnPath

"Returns an Array containing each object that has a nil value along the path."
<protected>

| array |
" gather the root objects with nil on the path "
array := Array new.
1 to: self size - 1 do: [ :i |
    self _addAllFor: nil
        into: array
        offset: i
        pathTerm: (self at: i)
].
^ array
%

category: 'Sorting'
method: RangeEqualityIndex
_sortAscendingOn: indexObjs directions: theBooleans

""
<protected>

| array cnt prevKey prevVal marker sorter btreeOnly offset pathTerm tmpArray 
  stream val |

prevKey := #_incompletePathTraversal.
prevVal := #_incompletePathTraversal.

" create the Array to hold the objects "
array := Array new: nscRoot size.

" build a sorter to be used for secondary sorts "
sorter := self _getSecondarySorterOn: indexObjs directions: theBooleans.

" start the insertion into the Array at the beginning "
self isIndexOnRootNsc
    ifTrue: [ cnt := 0 ]
    ifFalse: [ cnt := nscRoot _idxOccurrencesOf: nil ].

" gather objects with nil along the path "
tmpArray := self _getObjectsWithNilOnPath.
1 to: tmpArray size do: [ :i |
    cnt := cnt + 1.
    array at: cnt put: (tmpArray at: i)
].

btreeOnly := self size == 1.
" if no secondary sorts "
sorter == nil
  ifTrue: [
    " if path length of one "
    btreeOnly
      ifTrue: [
        btreeRoot _putAscendingValuesInto: array startingAt: cnt + 1.
      ]
      ifFalse: [
        offset := self lastPathComponentsDictionaryOffset max: 1.
        pathTerm := self at: offset.
        tmpArray := Array new.

        stream := BtreeReadStream on: btreeRoot.
        [ stream atEnd ] whileFalse: [
          val := stream _btreeNext.
         (val == prevVal)
            ifFalse: [
              self _addAllFor: val into: tmpArray offset: offset pathTerm: pathTerm.
              1 to: tmpArray size do: [ :i |
                cnt := cnt + 1.
                array at: cnt put: (tmpArray at: i)
              ].
              tmpArray size: 0
            ].
          prevVal := val
        ]
      ].
    ^ array
  ].

" must check for possibility of duplicate keys (there is a secondary sort) "
marker := cnt max: 1.
offset := self lastPathComponentsDictionaryOffset max: 1.
pathTerm := self at: offset.
tmpArray := Array new.

" Iterate over the key/values in the B-tree leaf nodes "
btreeRoot _leafKeysAndValuesDo: [ :key :value |

  btreeOnly
    ifTrue: [
      cnt := cnt + 1.
      array at: cnt put: value.
    ]
    ifFalse: [
     (value == prevVal)
        ifFalse: [
          self _addAllFor: value into: tmpArray offset: offset pathTerm: pathTerm.
          1 to: tmpArray size do: [ :i |
            cnt := cnt + 1.
            array at: cnt put: (tmpArray at: i)
          ].
          tmpArray size: 0
        ]
    ].

    " see if finished a run of duplicate keys "
    key ~= prevKey
      ifTrue: [
        marker < (cnt - 1)
          ifTrue: [ " sort on secondary keys "
            marker to: cnt - 1 do: [ :i | sorter addObject: (array at: i) ].
            sorter sortInto: array startingAt: marker.
            sorter clear
          ].
        marker := cnt
      ].
    prevKey := key.
    prevVal := value
].
" see if ending with duplicate keys "
( sorter ~~ nil _and: [ marker < cnt ] )
    ifTrue: [ " sort on secondary keys "
       marker to: cnt do: [ :i | sorter addObject: (array at: i) ].
       sorter sortInto: array startingAt: marker.
       sorter clear
    ].

^ array
%

category: 'Sorting'
method: RangeEqualityIndex
_sortDescendingOn: indexObjs directions: theBooleans

""
<protected>

| array cnt prevKey prevVal marker sorter btreeOnly offset pathTerm tmpArray 
  stream val |

prevKey := #_incompletePathTraversal.
prevVal := #_incompletePathTraversal.

" create the Array to hold the objects "
array := Array new: nscRoot size.

" build a sorter to be used for secondary sorts "
sorter := self _getSecondarySorterOn: indexObjs directions: theBooleans.

" start the insertion into the Array at the beginning "
self isIndexOnRootNsc
    ifTrue: [ cnt := array size  ]
    ifFalse: [ cnt := array size - (nscRoot _idxOccurrencesOf: nil) ].

" gather objects with nil along the path "
tmpArray := self _getObjectsWithNilOnPath.
1 to: tmpArray size do: [ :i |
    array at: cnt put: (tmpArray at: i).
    cnt := cnt - 1.
].

" if path length of one and no secondary sorts,
  optimize putting objects in the result Array "
btreeOnly := self size == 1.
sorter == nil
  ifTrue: [
    btreeOnly
      ifTrue: [
        btreeRoot _putDescendingValuesInto: array startingAt: cnt.
      ]
      ifFalse: [
        offset := self lastPathComponentsDictionaryOffset max: 1.
        pathTerm := self at: offset.
        tmpArray := Array new.

        stream := BtreeReadStream on: btreeRoot.
        [ stream atEnd ] whileFalse: [
          val := stream _btreeNext.

          (val == prevVal)
            ifFalse: [
              self _addAllFor: val into: tmpArray offset: offset pathTerm: pathTerm.
              1 to: tmpArray size do: [ :i |
                array at: cnt put: (tmpArray at: i).
                cnt := cnt - 1.
              ].
              tmpArray size: 0
            ].
          prevVal := val.
        ]
      ].
    ^ array
  ].

" iterate over the receiver looking for duplicate keys "
marker := cnt.
offset := self lastPathComponentsDictionaryOffset max: 1.
pathTerm := self at: offset.
tmpArray := Array new.

" iterate over the key/values in the B-tree leaf nodes "
btreeRoot _leafKeysAndValuesDo: [ :key :value |
  btreeOnly
    ifTrue: [ 
      array at: cnt put: value.
      cnt := cnt - 1.
    ]
    ifFalse: [
      (value == prevVal)
        ifFalse: [
          self _addAllFor: value into: tmpArray offset: offset pathTerm: pathTerm.
            1 to: tmpArray size do: [ :i |
              array at: cnt put: (tmpArray at: i).
              cnt := cnt - 1.
            ].
            tmpArray size: 0
        ]
    ].

    " see if finished a run of duplicate keys "
    key ~= prevKey
      ifTrue: [
        marker > (cnt + 1)
          ifTrue: [ " sort on secondary keys "
            cnt + 2 to: marker do: [ :i | sorter addObject: (array at: i) ].
            sorter sortInto: array startingAt: cnt + 2.
            sorter clear
          ].
        marker := cnt + 1
      ].
    prevKey := key.
    prevVal := value.
].

" see if ending with duplicate keys "
( sorter ~~ nil _and: [ marker > 1 ] )
  ifTrue: [ " sort on secondary keys "
   1 to: marker do: [ :i | sorter addObject: (array at: i) ].
   sorter sortInto: array startingAt: 1.
     sorter clear
    ].
^ array
%
category: 'Sorting'
method: RangeEqualityIndex
_sortOn: indexObjs directions: theBooleans

"Returns an Array of objects from the root NSC that is sorted according to the
 paths traversed by indexObjs (with the direction of the sort specified in
 theBooleans)."
<protected>

(theBooleans at: 1)
    ifTrue: [ ^ self _sortAscendingOn: indexObjs directions: theBooleans ]
    ifFalse: [ ^ self _sortDescendingOn: indexObjs directions: theBooleans ]
%

category: 'Query Support'
method: RangeEqualityIndex
_findAllValuesThroughSetsNotIdenticalTo: aKey

"Returns a set of all values that satisfy the query.  This method works
 when the path is through a set-valued instance variable."
<protected>

| tmpVals tmpHolder startPt pathTerm |
tmpHolder := NscBuilder for: nscRoot speciesForSelect new max: self sizeForNscBuilder.

tmpVals := btreeRoot _lastValuesNotIdenticalTo: aKey into: Array new.

startPt := self lastPathComponentsDictionaryOffset.
pathTerm := self at: startPt.

1 to: tmpVals size do: [ :i |
    self _addAllFor: (tmpVals _at: i)
        into: tmpHolder
        offset: startPt
        pathTerm: pathTerm
].

^ tmpHolder completeBag
%

category: 'Statistics'
method: RangeEqualityIndex
_statisticsInto: dict

"Puts statistical information into the given dictionary."
<protected>

| arr arr1 arr2 |
arr := dict at: #BtreeSpaceUtilization ifAbsent: [
  dict
    at: #BtreeSpaceUtilization
    put: (Array with: Array new with: Array new)
].

arr1 := arr at: 1.
arr2 := arr at: 2.

btreeRoot _preOrderDo: [ :node |
  arr1 add: node numElements.
  arr2 add: node class maxNumberOfElements.
].
%

category: 'Accessing'
method: RangeEqualityIndex
nwayMergeProgress

"Returns the progress count of the n-way merge during index creation.
 Returns nil if index creation is complete."

<protected>

btreeRoot isBtreeNode
  ifTrue: [ ^ nil ].

^ btreeRoot sortNodes progress
%

unprotectmethods
