!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: setpatheval.gs,v 1.7 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy:
!   SetValuedPathEvaluator, PathEvaluator, Array, SequenceableCollection,
!   Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods SetValuedPathEvaluator
removeallclassmethods SetValuedPathEvaluator
protectmethods

category: 'For Documentation Installation only'
classmethod: SetValuedPathEvaluator
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Instance methods for SetValuedPathEvaluator
category: 'Query Select'
method: SetValuedPathEvaluator
findAllValuesWithComparison: selector to: aValue

"Returns a set of all values that satisfy the query.  Must iterate through all
 elements in the NSC, traversing down the path for each object."
<protected>

| tmpHolder obj bag |
tmpHolder := NscBuilder for: nsc speciesForSelect new max: self sizeForNscBuilder.
(self _doNotIncludeNilsForOperator: selector)
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            obj := bag _at: i.
            self
                traverseNonNils: obj
                performing: selector
                with: aValue
                startingAt: 1
                rootObject: obj
                into: tmpHolder
                ]
    ]
    ifFalse: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            obj := bag _at: i.
            self
                traverse: obj
                performing: selector
                with: aValue
                startingAt: 1
                rootObject: obj
                into: tmpHolder
        ]
    ].
^ tmpHolder completeBag
%

category: 'Query Detect'
method: SetValuedPathEvaluator
findFirstValueEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

^ self findFirstValueWithComparison: #= to: aValue
%

category: 'Query Detect'
method: SetValuedPathEvaluator
findFirstValueGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

"Returns the first value that satisfies the query."
<protected>

^ self 
    findFirstValueWithComparison: #_idxGreaterThanAndLessThanValues: 
    to:  #[ val1, bool1, val2, bool2 ]
%

category: 'Query Detect'
method: SetValuedPathEvaluator
findFirstValueIdenticalTo: aValue

"Returns the first value that satisfies the query."
<protected>

^ self findFirstValueWithComparison: #== to: aValue
%

category: 'Query Detect'
method: SetValuedPathEvaluator
findFirstValueWithComparison: selector to: aValue

"Returns the first value that satisfies the query.  Must iterate through the
 elements in the NSC, traversing down the path for each object."
<protected>

| obj bag |

" if path includes set-valued instance variable term, must use recursive algorithm "
(self _doNotIncludeNilsForOperator: selector)
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            obj := bag _at: i.
            (self traverseNonNils: obj
                performing: selector
                with: aValue
                startingAt: 1)
                ifTrue: [ ^ obj ]
        ]
    ]
    ifFalse: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            obj := bag _at: i.
            (self traverse: obj
                performing: selector
                with: aValue
                startingAt: 1)
                ifTrue: [ ^ obj ]
        ]
    ].
^ #_incompletePathTraversal
%

category: 'Testing'
method: SetValuedPathEvaluator
hasSetValuedTerm

"Returns true if the path has a term that indicates a set-valued instance
 variable."
<unprotected>

^ true
%

category: 'Traversing'
method: SetValuedPathEvaluator
traverse: anObject
performing: selector
with: argument
startingAt: offset

"Traverse the sub-objects of anObject, using the path names of the receiver
 starting at the path name at the given offset.  For each object at the end of
 the traversal, send the given comparison selector.  Returns the results of the
 comparison."
<protected>

| nextObj ivOffset pathName nscOffset sz |

nextObj := anObject.
sz := self size.
offset to: sz do: [ :i |

  pathName := self at: i.
  " if indexing over a set-valued instance variable "
  pathName = #*
    ifTrue: [
      nscOffset := i + 1.
      " for each object in the set-valued instance variable "
      nextObj do: [ :obj |
        " make recursive call "
        (self
          traverse: obj
          performing: selector
          with: argument
          startingAt: nscOffset)
            ifTrue: [ ^ true ]
      ].
      ^ false
    ]
    ifFalse: [
      nil == nextObj
        ifTrue: [ ^ false ].
      " get the next object along the path "
      ivOffset := nextObj class _ivOffsetOf: pathName.
      ivOffset == nil
        ifTrue: [ nextObj _errorInvalidOffset: pathName .
                  nextObj := nil
                ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .
      ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [ ^ false ]
    ]
].
^ nextObj perform: selector with: argument
%

category: 'Traversing'
method: SetValuedPathEvaluator
traverse: anObject
performing: selector
with: argument
startingAt: offset
rootObject: rootObject
into: resultSet

"Traverse the sub-objects of anObject, using the path names of the receiver
 starting at the path name at the given offset.  For each object at the end of
 the traversal, send the given comparison selector, and if the result is true,
 place the root object in a result set.  Returns the result set."
<protected>

| nextObj ivOffset pathName nscOffset sz |

nextObj := anObject.
sz := self size.
offset to: sz do: [ :i |
  pathName := self at: i.
  " if indexing over a set-valued instance variable "
  pathName = #*
    ifTrue: [
      nscOffset := i + 1.
      " for each object in the set-valued instance variable "
      nextObj do: [ :obj |
        " make recursive call "
        self
          traverse: obj
          performing: selector
          with: argument
          startingAt: nscOffset
          rootObject: rootObject
          into: resultSet
      ].
      ^ resultSet
    ]
    ifFalse: [
      nil == nextObj
        ifTrue: [ ^ resultSet ].
      " get the next object along the path "
      ivOffset := nextObj class _ivOffsetOf: pathName.
      ivOffset == nil
        ifTrue: [ nextObj _errorInvalidOffset: pathName  .
                  nextObj := nil ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ] .
      " see if nil was reached before the end of the path "
      ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [ ^ resultSet ]
    ]
].
" perform the selector and if the results are true, add the root object to the result set "
(nextObj perform: selector with: argument)
  ifTrue: [ resultSet add: rootObject ].

^ resultSet
%

category: 'Traversing'
method: SetValuedPathEvaluator
traverse: anObject startingAt: offset into: resultSet

"Traverse the sub-objects of all the objects in the NSC, using the path names
 of the receiver starting at the path term at the given offset.  Add each
 object at the end of the traversal to the result set.  Returns the result
 set."
<protected>

| nextObj ivOffset pathName nscOffset sz |

( nil == anObject _and: [ self isIndexOnNscElements not ] )
  ifTrue: [ ^ resultSet ].

nextObj := anObject.
sz := self size.
offset to: sz do: [ :i |

  pathName := self at: i.
  " if indexing over a set-valued instance variable "
  pathName = #*
    ifTrue: [
      nscOffset := i + 1.
      " for each object in the set-valued instance variable "
      nextObj do: [ :obj |
        " make recursive call "
        self
          traverse: obj
          startingAt: nscOffset
          into: resultSet
      ].
      ^ resultSet
    ]
    ifFalse: [
      nil == nextObj
          ifTrue: [ ^ resultSet ].
      " get the next object along the path "
      ivOffset := nextObj class _ivOffsetOf: pathName.
      ivOffset == nil
        ifTrue: [ nextObj _errorInvalidOffset: pathName .
                  nextObj := nil ]
        ifFalse: [ nextObj := nextObj instVarAt: ivOffset ].
    ].

  ( nil == nextObj _and: [ i ~= sz ] )
    ifTrue: [ ^ resultSet ]
].
resultSet add: nextObj.

^ resultSet
%

category: 'Traversing'
method: SetValuedPathEvaluator
traverseNonNils: anObject
performing: selector
with: argument
startingAt: offset

"Traverse the sub-objects of anObject, using the path names of the receiver
 starting at the path name at the given offset.  For each non-nil object at the
 end of the traversal, send the given comparison selector.  Returns the results
 of the comparison."
<protected>

| nextObj ivOffset pathName nscOffset sz |

nextObj := anObject.
sz := self size.
offset to: sz do: [ :i |

  pathName := self at: i.
  " if indexing over a set-valued instance variable "
  pathName = #*
    ifTrue: [
      nscOffset := i + 1.
      " for each object in the set-valued instance variable "
      nextObj do: [ :obj |
        " make recursive call "
        (self
          traverseNonNils: obj
          performing: selector
          with: argument
          startingAt: nscOffset)
            ifTrue: [ ^ true ]
      ].
      ^ false
    ]
    ifFalse: [
      nil == nextObj
        ifTrue: [ ^ false ].
      " get the next object along the path "
      ivOffset := nextObj class _ivOffsetOf: pathName.
      ivOffset == nil
        ifTrue:[ nextObj _errorInvalidOffset: pathName.  nextObj := nil ]
        ifFalse:[ nextObj := nextObj instVarAt: ivOffset ].

      ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [ ^ false ]
    ]
].
nil == nextObj
  ifTrue: [ ^ false ]
  ifFalse: [ ^ nextObj perform: selector with: argument ]
%

category: 'Traversing'
method: SetValuedPathEvaluator
traverseNonNils: anObject
performing: selector
with: argument
startingAt: offset
rootObject: rootObject
into: resultSet

"Traverse the sub-objects of anObject, using the path names of the receiver]
 starting at the path name at the given offset.  For each non-nil object at the
 end of the traversal, send the given comparison selector, and if the result is
 true, place the root object in a result set.  Returns the result set."
<protected>

| nextObj ivOffset pathName nscOffset sz |

nextObj := anObject.
sz := self size.
offset to: sz do: [ :i |
  pathName := self at: i.
  " if indexing over a set-valued instance variable "
  pathName = #*
    ifTrue: [
      nscOffset := i + 1.
      " for each object in the set-valued instance variable "
      nextObj do: [ :obj |
        " make recursive call "
        self
          traverseNonNils: obj
          performing: selector
          with: argument
          startingAt: nscOffset
          rootObject: rootObject
          into: resultSet
      ].
      ^ resultSet
    ]
    ifFalse: [
      nil == nextObj
        ifTrue: [ ^ resultSet ].
      " get the next object along the path "
      ivOffset := nextObj class _ivOffsetOf: pathName.
      ivOffset == nil
       ifTrue:[nextObj _errorInvalidOffset: pathName.  nextObj := nil ]
       ifFalse:[nextObj := nextObj instVarAt: ivOffset ] .

      " see if nil was reached before the end of the path "
      ( nil == nextObj _and: [ i ~= sz ] )
        ifTrue: [ ^ resultSet ]
    ]
].
" perform the selector and if the results are true, add the root object to the result set "
(nil ~~ nextObj _and: [ nextObj perform: selector with: argument ] )
  ifTrue: [ resultSet add: rootObject ].

^ resultSet
%

category: 'Query Support'
method: SetValuedPathEvaluator
traverse: anObject andCompare: selector with: aValue

"Traverse the given object and perform the comparison operation on the object at
 the end of the path and the given value."
<protected>

^ (self _doNotIncludeNilsForOperator: selector)
    ifTrue: [
        self traverseNonNils: anObject
            performing: selector
            with: aValue
            startingAt: 1
    ]
    ifFalse: [
        self traverse: anObject
            performing: selector
            with: aValue
            startingAt: 1
    ]
%

category: 'Query Support'
method: SetValuedPathEvaluator
traverse: anObject andCompare: selector withTraverser: aTraverser

"Traverse the given object and perform the comparison operation on the object at
 the end of the path and the value obtained by using the traverser."
<protected>

| tmp |

aTraverser hasSetValuedTerm
    ifTrue: [ " both are set-valued paths "
        tmp := Array new.
        aTraverser traverse: anObject startingAt: 1 into: tmp.
        1 to: tmp size do: [ :i |
            (self traverse: anObject
                performing: selector
                with: (tmp _at: i)
                startingAt: 1)
                ifTrue: [ ^ true ]
        ].
        ^ false
    ]
    ifFalse: [
        ^ self traverse: anObject
            performing: selector
            with: (aTraverser traverseObject: anObject)
            startingAt: 1
    ]
%

category: 'Testing'
method: SetValuedPathEvaluator
isIndexOnNscElements

"Returns whether the receiver is a path directly on the elements of an NSC
 (either the root NSC or a set-valued instance variable along the path)."
<protected>

^ (self at: self size) = #*
%

unprotectmethods
